import os, gc
import argparse
import math

import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader

from .dataset import get_dataset
from .model import get_model
from .util.misc import get_file_name, str_localtime, seed_everything
from .util.hook_features import hook_features
from .util.yaml_config import get_config, read_yaml, write_yaml
from .util.postprocess import postprocess, apply_constraints
from .util.data_sampler import DeviceMultiDataLoader
from .util.RNA_kit import read_SS, write_SS, read_fasta, connects2dbn, mat2connects, compute_confidence, remove_lone_pairs, merge_connects


SRC_DIR = os.path.dirname(os.path.realpath(__file__))


def load_eval_checkpoints(ckpt_dir, RNA_model, model_opts, device, ckpt_names=None):
    if not os.path.exists(ckpt_dir):
        raise Exception(f'[Error] Checkpoint directory not exist: {ckpt_dir}')
    models = []
    if ckpt_names is None:
        ckpt_names = sorted(os.listdir(ckpt_dir))
    for ckpt_name in ckpt_names:
        ckpt_path = os.path.join(ckpt_dir, ckpt_name)
        print(f'Loading {os.path.abspath(ckpt_path)}')
        model = RNA_model(**model_opts)
        model = model.to(device)
        model.load_state_dict(torch.load(ckpt_path, map_location=torch.device('cpu'), weights_only=True))
        model.eval()
        models.append(model)
    if models == []:
        raise Exception(f'[Error] No checkpoint found in {ckpt_dir}')
    return models


class BPfold_predict:
    def __init__(self, checkpoint_dir, config_file=None):
        '''
        Init

        Parameters
        ----------
        checkpoint_dir: str
            Directory of checkpoints that contain trained parameters.
        config_file: str
            if None, default: os.path.join(os.path.dirname(__file__), 'configs/config.yaml')
        '''
        self.tmp_dir = '.BPfold_tmp_files'
        self.para_dir = os.path.join(SRC_DIR, 'paras')
        config_file = os.path.join(SRC_DIR, 'configs/config.yaml')
        opts = get_config(config_file)
        model_name = opts['model']['model_name']
        data_name = opts['dataset']['data_name']
        data_opts = opts['dataset'][data_name]
        model_opts = opts['model'][model_name]
        common_opts = opts['common']
        data_opts.update(common_opts)
        model_opts.update(common_opts)

        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        # load checkpoints
        RNA_model = get_model(model_name)
        self.models = load_eval_checkpoints(checkpoint_dir, RNA_model, model_opts, self.device)

        self.data_opts = data_opts

    def predict(self, input_seqs=None, input_path=None, batch_size=1, num_workers=1, hook_features=False, save_contact_map=False, ignore_nc=False):
        '''
        BPfold `predict function`, specify input_seqs or input_path 

        Parameters
        ----------
        input_seqs: [str]
            RNA seq list
        input_path: str
            fasta path containing multi RNA seqs or Secondary structure path in format of 'dbn', 'bpseq' or 'ct'.

        Returns
        -------
        ret: dict
            str,      str, [int],    float
            seq_name, seq, connects, CI
        '''
        dl = self.get_predict_loader(self.data_opts, self.device, input_seqs, input_path, batch_size, num_workers, data_name='RNAseq')
        for data_dic, _ in dl:
            with torch.no_grad(): #,torch.amp.autocast(self.device):
                # torch.nan_to_num
                # BS x forward_batch_Lmax x forward_batch_Lmax
                pred_batch = torch.stack([model(data_dic) for model in self.models], 0).mean(0)

                # remove `begin` and `end` tokens
                forward_batch_Lmax = data_dic['forward_mask'].sum(-1).max()
                batch_Lmax = forward_batch_Lmax-2
                pred_batch = pred_batch[:, 1:batch_Lmax+1, 1:batch_Lmax+1]
                seq_onehot = data_dic['seq_onehot'][:, 1:batch_Lmax+1, :]
                nc_map = data_dic['nc_map'][:, 1:batch_Lmax+1, 1:batch_Lmax+1]
                masks = data_dic['mask'][:, 1:batch_Lmax+1, 1:batch_Lmax+1]
                seqs = data_dic['ori_seq']
                names = data_dic['name']

                if hook_features:
                    # hook_dir
                    hook_dir = os.path.join(self.tmp_dir, 'hook_features')
                    os.makedirs(hook_dir, exist_ok=True)
                    hook_module_names = ['TransformerEncoderLayer', 'ResConv2dSimple']
                    hooker = hook_features(self.models[0], hook_module_names)
                    module_count = {}
                    for module_name, input_feature, output_feature in zip(*hooker.get_hook_results()):
                        if module_name not in module_count:
                            module_count[module_name] = 0
                        module_count[module_name]+=1
                        save_name = f'{names[0]}_{module_name}_{module_count[module_name]:02d}'
                        save_path = os.path.join(hook_dir, save_name + '.npy')
                        out_map = output_feature[0].detach().cpu().numpy()
                        np.save(save_path, out_map)

                # postprocess
                ret_pred, ret_pred_nc, _, _ = postprocess(pred_batch, seq_onehot, nc_map, return_score=False, return_nc=True)

                # save pred
                for i in range(len(ret_pred)):
                    length = len(seqs[i])
                    seq_name = names[i]
                    mat = pred_batch[i][masks[i]].reshape(length, length).detach().cpu().numpy()
                    mat_post = ret_pred[i][masks[i]].reshape(length, length).detach().cpu().numpy()
                    CI = compute_confidence(mat, mat_post)
                    connects = mat2connects(mat_post)
                    connects = remove_lone_pairs(connects)

                    ## save contact maps before and after postprocessing
                    if save_contact_map:
                        save_data_dir = os.path.join(self.tmp_dir, 'contact_map')
                        os.makedirs(save_data_dir, exist_ok=True)
                        ## save numpy arr, before/after postprocessing
                        mat = pred_batch[i][masks[i]].reshape(length, length).detach().cpu().numpy()
                        mat_post = ret_pred[i][masks[i]].reshape(length, length).detach().cpu().numpy()
                        np.save(os.path.join(save_data_dir, f'{seq_name}.npy'), mat)
                        np.save(os.path.join(save_data_dir, f'{seq_name}_post.npy'), mat_post)
                    results = {'seq_name': seq_name, 'seq': seqs[i], 'connects': connects, 'CI': CI}

                    ## NC pairs
                    if not ignore_nc: # not accurate enough, to be improved
                        mat_nc_post = ret_pred_nc[i][masks[i]].reshape(length, length).detach().cpu().numpy()
                        results['connects_nc'] = mat2connects(mat_nc_post)

                    yield results

                # clean memory
                del data_dic, pred_batch, seq_onehot, nc_map, masks
                gc.collect()
                torch.cuda.empty_cache()


    def gen_info_dic(self, input_seqs, input_path, data_name='RNAseq'):
        def valid_seq(seq):
            return seq.isalpha()
        def process_one_seq(name, seq, src_fasta_path=''):
            if valid_seq(seq):
                return {'seq': seq, 'name': name, 'length': len(seq), 'dataset': data_name}
            else:
                print(f'[Warning] Invalid seq, containing non-alphabet character, ignored: {src_fasta_path}{name}="{seq}"')
        def process_one_file(file_path, data_name='RNAseq'):
            file_name, suf = get_file_name(file_path, return_suf=True)
            if suf.lower() in {'.fasta', '.fa'}: # fasta file
                for name, seq in read_fasta(file_path):
                    info_dic = process_one_seq(name, seq, file_path+': ')
                    if info_dic is not None:
                        yield info_dic
            elif suf.lower() in {'.dbn', '.ct', '.bpseq'}: # SS file
                seq, _ = read_SS(file_path)
                yield {'path': file_path, 'length': len(seq), 'dataset': data_name}
            else: # unknown filetype
                print(f'[Warning] Unknown file type, ignored: {file_path}')

        if input_seqs:
            if isinstance(input_seqs, str):
                input_seqs = [input_seqs]
            time_str = str_localtime()
            for idx, seq in enumerate(input_seqs):
                seq_name = f'seq_{time_str}_{idx+1}'
                info_dic = process_one_seq(seq_name, seq)
                if info_dic:
                    yield info_dic
        if input_path:
            if os.path.isfile(input_path):
                yield from process_one_file(input_path, data_name)
            else:
                for pre, ds, fs in os.walk(input_path):
                    for f in fs:
                        yield from process_one_file(os.path.join(pre, f), data_name)

    def get_predict_loader(self, data_opts, device, input_seqs, input_path, batch_size, num_workers, data_name='RNAseq'):
        data_class = get_dataset(data_name)
        data_opts['predict_files'] = list(self.gen_info_dic(input_seqs, input_path, data_name) )
        ds = data_class(phase='predict', verbose=False, para_dir=self.para_dir, **data_opts)
        dl = torch.utils.data.DataLoader(ds, batch_size=batch_size, shuffle=False, drop_last=False, num_workers=num_workers)
        return DeviceMultiDataLoader([dl], device, keywords=ds.to_device_keywords)

    def save_pred_results(self, pred_results, save_dir='.', save_name=None, out_type='csv', hide_dbn=False):
        '''
        Save pred_results predicted by `BPfold_predict.predict` in `save_dir` in format of `out_type`.

        Parameters
        ----------
        pred_results: Iterable(dict)
            Each item is a dict which contains keys `seq_name`, `seq`, `connects` and `CI` (optional: connects_nc).
        save_dir: str
            save_dir, if not exist, will mkdir.
        save_name: str
            if not available, use basename of save_dir
        out_type: str
            csv, bpseq, ct, dbn
        '''
        def print_result(save_name, idx, seq, dbn, CI, hide_dbn=False, num_digit=7):
            CI_str = f'CI={CI:.3f}' if CI>=0.3 else 'CI<0.3'
            print(f"[{str(idx).rjust(num_digit)}] saved in \"{save_name}\", {CI_str}")
            if not hide_dbn:
                print(f'{seq}\n{dbn}')

        os.makedirs(save_dir, exist_ok=True)
        if out_type=='csv':
            if save_name is None:
                save_name = os.path.basename(os.path.abspath(save_dir))
            df = pd.DataFrame(pred_results)
            df['dbn'] = df['connects'].apply(connects2dbn)
            csv_path = os.path.join(save_dir, f'{save_name}.csv')
            num_digit = math.ceil(math.log(len(df), 10))
            for idx, row in enumerate(df.itertuples()):
                print_result(f'{csv_path}:line{idx+2}:{row.seq_name}', idx+1, row.seq, row.dbn, row.CI, hide_dbn=hide_dbn, num_digit=num_digit)
            ## NC pairs
            if 'connects_nc' in df.columns: # not accurate, to be improved
                df['dbn_nc'] = df['connects_nc'].apply(connects2dbn)
                df['connects_mix'] = df.apply(lambda row: merge_connects(row['connects'], row['connects_nc']), axis=1)
                # df['connects_mix'] = merge_connects(df['connects'], df['connects_nc']) # Error, 函数必须能够向量化操作才行
                df['dbn_mix'] = df['connects_mix'].apply(connects2dbn)
            df.to_csv(csv_path, index=False)
            print(f"Predicted structures in format of dot-bracket are saved in \"{csv_path}\".")

        else:
            num_digit = 7
            confidence_dic = {}
            for ct, res_dic in enumerate(pred_results):
                seq_name = res_dic['seq_name']
                seq = res_dic['seq']
                connects = res_dic['connects']
                CI = res_dic['CI']
                confidence_dic[seq_name] = CI
                path = os.path.join(save_dir, seq_name+f'.{out_type}')
                write_SS(path, seq, connects)
                print_result(path, ct+1, seq, connects2dbn(connects), CI, hide_dbn=hide_dbn, num_digit=num_digit)

                ## NC pairs
                if 'connects_nc' in res_dic: # not accurate, to be improved
                    connects_nc = res_dic['connects_nc']
                    path = os.path.join(save_dir, seq_name+'_nc'+f'.{out_type}')
                    write_SS(path, seq, connects_nc)
                    connects_mix = merge_connects(connects, connects_nc)
                    path = os.path.join(save_dir, seq_name+'_mix'+f'.{out_type}')
                    write_SS(path, seq, connects_mix)
                    if not hide_dbn:
                        print(f'{connects2dbn(connects_nc)} NC')
                        print(f'{connects2dbn(connects_mix)} MIX')
            confidence_path = os.path.join(os.path.dirname(save_dir), os.path.basename(save_dir)+f'_confidence_{str_localtime()}.yaml')
            if confidence_dic:
                print(f"Confidence indexes are saved in \"{confidence_path}\"")
                write_yaml(confidence_path, confidence_dic)


def show_examples():
    method_name = 'BPfold'
    print('Please specify "--seq" or "--input" argument for input sequences or input file. Such as:')
    print(f'$ {method_name} --checkpoint_dir PATH_TO_CHECKPOINT --seq GGUAAAACAGCCUGU AGUAGGAUGUAUAUG --output {method_name}_results')
    print(f'$ {method_name} --checkpoint_dir PATH_TO_CHECKPOINT --input examples/examples.fasta # (multiple sequences are supported)')
    print(f'$ {method_name} --checkpoint_dir PATH_TO_CHECKPOINT --input examples/URS0000D6831E_12908_1-117.bpseq # .bpseq, .ct, .dbn')
    exit()


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--checkpoint_dir', type=str, help='Directory of checkpoints that contain trained parameters.', required=True)
    parser.add_argument('-s', '--seq', nargs='*', help='RNA sequences')
    parser.add_argument('-i', '--input', type=str, help='Input fasta file or directory which contains fasta files, supporting multiple seqs and multiple formats, such as fasta, bpseq, ct or dbn.')
    parser.add_argument('-o', '--output', type=str, default='BPfold_results', help='output directory')
    parser.add_argument('-g', '--gpu', type=str, default='0')
    parser.add_argument('--out_type', default='csv', choices=['bpseq', 'ct', 'dbn', 'csv'], help='Saved file type.')
    parser.add_argument('--batch_size', type=int, default=1)
    parser.add_argument('--num_workers', type=int, default=1)
    parser.add_argument('--hide_dbn', action='store_true', help='Once specified, the output sequence and predicted DBN won\'t be printed.')
    parser.add_argument('--ignore_nc', action='store_true', help='Ignore non-canonical pairs when saving predictions.')
    parser.add_argument('--save_contact_map', action='store_true')
    parser.add_argument('--hook_features', action='store_true')
    args = parser.parse_args()
    return args


def main():
    print('>> Welcome to use "BPfold" for predicting RNA secondary structure!')
    args = parse_args() 
    os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu
    seed_everything(42)

    # usage
    if args.input is None and args.seq is None:
        show_examples()

    BPfold_predictor = BPfold_predict(checkpoint_dir=args.checkpoint_dir)
    pred_results = BPfold_predictor.predict(args.seq, args.input, args.batch_size, args.num_workers, args.hook_features, args.save_contact_map, ignore_nc=args.ignore_nc)
    save_name = get_file_name(args.input) if args.input else None
    BPfold_predictor.save_pred_results(pred_results, save_dir=args.output, save_name=save_name, out_type=args.out_type)
    del BPfold_predictor
    gc.collect()
    torch.cuda.empty_cache()
    print('Program Finished!')


if __name__ == '__main__':
    main()



import torch
import torch.nn.functional as F


def get_base_index():
    return {'A': 0, 'U': 1, 'C': 2, 'G': 3}


def constraint_matrix_batch(x, loop_min_len=2, is_nc=False):
    base_index = get_base_index()
    base_a = x[:, :, base_index['A']]
    base_u = x[:, :, base_index['U']]
    base_c = x[:, :, base_index['C']]
    base_g = x[:, :, base_index['G']]
    batch = base_a.shape[0]
    length = base_a.shape[1]

    # canonical pairs
    au = torch.matmul(base_a.view(batch, length, 1), base_u.view(batch, 1, length))
    au_ua = au + torch.transpose(au, -1, -2)
    cg = torch.matmul(base_c.view(batch, length, 1), base_g.view(batch, 1, length))
    cg_gc = cg + torch.transpose(cg, -1, -2)
    ug = torch.matmul(base_u.view(batch, length, 1), base_g.view(batch, 1, length))
    ug_gu = ug + torch.transpose(ug, -1, -2)
    ret = au_ua + cg_gc + ug_gu # batch x L x L

    ## non-canonical pairs
    if is_nc:
        ac = torch.matmul(base_a.view(batch, length, 1), base_c.view(batch, 1, length))
        ac_ca = ac + torch.transpose(ac, -1, -2)
        ag = torch.matmul(base_a.view(batch, length, 1), base_g.view(batch, 1, length))
        ag_ga = ag + torch.transpose(ag, -1, -2)
        uc = torch.matmul(base_u.view(batch, length, 1), base_c.view(batch, 1, length))
        uc_cu = uc + torch.transpose(uc, -1, -2)
        aa = torch.matmul(base_a.view(batch, length, 1), base_a.view(batch, 1, length))
        uu = torch.matmul(base_u.view(batch, length, 1), base_u.view(batch, 1, length))
        cc = torch.matmul(base_c.view(batch, length, 1), base_c.view(batch, 1, length))
        gg = torch.matmul(base_g.view(batch, length, 1), base_g.view(batch, 1, length))
        ret += ac_ca + ag_ga + uc_cu + aa + uu + cc + gg

    # remove sharp loop 
    for b in range(batch):
        for i in range(length):
            for j in range(i, length):
                if j-i<loop_min_len+1: # i [i+1, i+2...] i+loop_len+1=j:  [] reprs loop region
                    ret[b, i, j] = ret[b, j, i] = 0
    return ret


def contact_a(a_hat, m):
    a = a_hat * a_hat
    a = (a + torch.transpose(a, -1, -2)) / 2
    a = a * m
    return a


def sign(x):
    return (x > 0).type(x.dtype)


def soft_sign(x):
    k = 1
    return 1.0/(1.0+torch.exp(-2*k*x))


def apply_constraints(u, x, lr_min, lr_max, num_itr, rho=0.0, with_l1=False, s=2.2, is_nc=False):
    """
    :param u: utility matrix, u is assumed to be symmetric, in batch
    :param x: RNA sequence, in batch
    :param lr_min: learning rate for minimization step
    :param lr_max: learning rate for maximization step (for lagrangian multiplier)
    :param num_itr: number of iterations
    :param rho: sparsity coefficient
    :param with_l1:
    :return:
    """
    m = constraint_matrix_batch(x, is_nc=is_nc).float()
    # u with threshold
    # equivalent to sigmoid(u) > 1/(e^(-2.2)+1)
    # u = (u > 2.2).type(torch.FloatTensor) * u
    u = soft_sign(u - s) * u

    # initialization
    a_hat = (torch.sigmoid(u)) * soft_sign(u - s).detach()
    lmbd = F.relu(torch.sum(contact_a(a_hat, m), dim=-1) - 1).detach()

    # gradient descent
    for t in range(num_itr):
        grad_a = (lmbd * soft_sign(torch.sum(contact_a(a_hat, m), dim=-1) - 1)).unsqueeze_(-1).expand(u.shape) - u / 2
        grad = a_hat * m * (grad_a + torch.transpose(grad_a, -1, -2))
        a_hat -= lr_min * grad
        lr_min = lr_min * 0.99

        if with_l1:
            a_hat = F.relu(torch.abs(a_hat) - rho * lr_min)

        lmbd_grad = F.relu(torch.sum(contact_a(a_hat, m), dim=-1) - 1)
        lmbd += lr_max * lmbd_grad
        lr_max = lr_max * 0.99
    out_a = contact_a(a_hat, m)
    return out_a


def postprocess(pred_batch, seq_onehot, nc_map, return_nc=False, return_score=False):
    ''' nc: non-canonical '''
    def get_mat_score(pred_ori, pred_post):
        ret = []
        for row_ori, row_post in zip(pred_ori, pred_post):
            val = row_ori[row_post.argmax()]
            new_row = row_ori[:]
            new_row[new_row>val] = new_row.min()
            ret.append(new_row)
        return ret

    ret_pred = []
    ret_pred_nc = []
    ret_score = []
    ret_score_nc = []
    for i in range(pred_batch.shape[0]):
        pred = apply_constraints(pred_batch[i:i+1], seq_onehot[i:i+1], 0.01, 0.1, 100, 1.6, True, 1.5)
        pred = (pred > 0.5).float()
        ret_pred.append(pred[0])

        if return_score:
            ret_score.append(get_mat_score(pred_batch[i], pred[0]))
        if return_nc:
            pred_nc = apply_constraints(pred_batch[i:i+1], seq_onehot[i:i+1], 0.01, 0.1, 100, 0.5, True, 0.5, is_nc=True)
            pred_nc =  nc_map * pred_nc
            pred_nc = (pred_nc > 0.5).float()
            ret_pred_nc.append(pred_nc[0])
            if return_score:
                ret_score_nc.append(get_mat_score(pred_batch[i], pred_nc[0]))
    return ret_pred, ret_pred_nc, ret_score, ret_score_nc



class hook_features:
    def __init__(self, model, hook_module_names):
        '''
        Parameters
        ----------
        model: pytorch model
        hook_module_names: [str]
            names of modules to be hooked
        '''
        self.myinit()
        self.register_hook(model, hook_module_names)

    def myinit(self):
        self.module_names = []
        self.input_features =  []
        self.output_features =  []

    def register_hook(self, model, hook_module_names):
        for module in model.modules():  # model.children() # only return the direct children
            if self.get_module_name(module) in hook_module_names:
                module.register_forward_hook(hook=self.hook_func)

    def get_module_name(self, module):
        '''
            module: pytorch module
        '''
        return str(module.__class__)[:-2].split('.')[-1] 

    def hook_func(self, module, in_feat, out_feat):
        self.module_names.append(self.get_module_name(module))
        self.input_features.append(in_feat)
        self.output_features.append(out_feat)

    def get_hook_results(self):
        '''
            This function should be called after calling `model.forward`.
        '''
        names = self.module_names[:]
        inputs = self.input_features[:]
        outputs = self.output_features[:]
        self.myinit()
        return names, inputs, outputs



__all__ = ["get_config", "update_config", "getEasyDict", "read_yaml", "toYaml", "write_yaml"]

import yaml


def getEasyDict(dic):
    new_dic = dic
    for k, v in dic.items():
        if isinstance(v, dict):
            new_dic[k] = getEasyDict(v)
        else:
            new_dic[k] = v
    return EasyDict(new_dic)


class EasyDict(object):
    def __init__(self, opt):
        self.opt = opt

    def __getattribute__(self, name):
        if name == 'opt' or name.startswith("_") or name not in self.opt:
            return object.__getattribute__(self, name)
        else:
            return self.opt[name]

    def __setattr__(self, name, value):
        if name == 'opt':
            object.__setattr__(self, name, value)
        else:
            self.opt[name] = value

    def __getitem__(self, name):
        return self.opt[name]

    def __setitem__(self, name, value):
        self.opt[name] = value

    def __contains__(self, item):
        return item in self.opt

    def __repr__(self):
        return self.opt.__repr__()

    def __len__(self):
        return len(self.opt)

    def keys(self):
        return self.opt.keys()

    def values(self):
        return self.opt.values()

    def items(self):
        return self.opt.items()

    def update(self, dic):
        for k, v in dic.items():
            self.opt[k] = v


def resolve_expression(config):
    if type(config) is dict:
        new_config = {}
        for k, v in config.items():
            if type(v) is str and v.startswith("!!python"):
                v = eval(v[8:])
            elif type(v) is dict:
                v = resolve_expression(v)
            new_config[k] = v
        config = new_config
    return config


def get_config(config_file, config_names=[]):
    ''' load config from file
    '''
    config = resolve_expression(read_yaml(config_file))
    if type(config_names) == str:
        return getEasyDict(config[config_names])
    while len(config_names) != 0:
        config_name = config_names.pop(0)
        if config_name not in config:
            raise ValueError("Invalid config name: {}".format(config_name))
        config = config[config_name]

    return getEasyDict(config)


def update_config(config, args):
    ''' rewrite default config with user input
    '''
    if args is None:
        return
    if hasattr(args, "__dict__"):
        args = args.__dict__
    for arg, val in args.items():
        # if not (val is None or val is False) and arg in config: config[arg] = val
        # TO FIX: this may cause bugs for other programs
        if arg in config and val is not None:
            config[arg] = val

    for _, val in config.items():
        if isinstance(val, dict) or isinstance(val, EasyDict):
            update_config(val, args)


def resolve_obj(obj):
    if isinstance(obj, list):
        return [resolve_obj(i) for i in obj]
    elif isinstance(obj, dict) or isinstance(obj, EasyDict):
        return {k: resolve_obj(v) for k, v in obj.items()}
    elif isinstance(obj, tuple):
        return list(obj)
    else:
        return obj


def read_yaml(path):
    with open(path) as fp:
        return yaml.load(fp, Loader=yaml.FullLoader)


def toYaml(path, dic, listOneLine=True):
    dic = resolve_obj(dic)
    if listOneLine:
        data_str = yaml.dump(dic)
    else:
        data_str = yaml.dump(dic, default_flow_style=False)
    with open(path, 'w') as f:
        f.write(data_str)
write_yaml = toYaml



from typing import ClassVar

import numpy as np
import torch
import torch.nn as nn

from .pos_embed import *


class SE_Block(nn.Module):
    "credits: https://github.com/moskomule/senet.pytorch/blob/master/senet/se_module.py#L4"
    def __init__(self, c, r=1):
        super().__init__()
        self.squeeze = nn.AdaptiveAvgPool2d(1)
        self.excitation = nn.Sequential(
            nn.Linear(c, c // r, bias=False),
            nn.ReLU(inplace=True),
            nn.Linear(c // r, c, bias=False),
            nn.Sigmoid()
        )

    def forward(self, x):
        bs, c, _, _ = x.shape
        y = self.squeeze(x).view(bs, c)
        y = self.excitation(y).view(bs, c, 1, 1)
        return x * y.expand_as(x)          
        
class ResConv2dSimple(nn.Module):
    def __init__(self, 
                 in_c, 
                 out_c,
                 kernel_size=7,
                 use_se = False,
                ):  
        super().__init__()
        if use_se:
            self.conv = nn.Sequential(
                # b c w h
                nn.Conv2d(in_c,
                          out_c, 
                          kernel_size=kernel_size, 
                          padding="same", 
                          bias=False),
                # b w h c#
                nn.BatchNorm2d(out_c),
                SE_Block(out_c),
                nn.GELU(),
                # b c e 
            )
            
        else:
            self.conv = nn.Sequential(
                # b c w h
                nn.Conv2d(in_c,
                          out_c, 
                          kernel_size=kernel_size, 
                          padding="same", 
                          bias=False),
                # b w h c#
                nn.BatchNorm2d(out_c),
                nn.GELU(),
                # b c e 
            )
        
        if in_c == out_c:
            self.res = nn.Identity()
        else:
            self.res = nn.Sequential(
                nn.Conv2d(in_c,
                          out_c, 
                          kernel_size=1, 
                          bias=False)
            )

    def forward(self, x):
        # b e s 
        h = self.conv(x)
        x = self.res(x) + h
        return x
    

class MultiHeadSelfAttention(nn.Module):
    def __init__(self,
                 hidden_dim: int,
                 positional_embedding: str,
                 num_heads: int = None,
                # k_dim: int = None,
                # v_dim: int = None,
                 dropout: float = 0.10, 
                 bias: bool = True,
                 temperature: float = 1,
                 use_se = False,
                ):
        super().__init__()
        
        assert positional_embedding in ("dyn", "alibi")
        self.positional_embedding = positional_embedding
        self.hidden_dim = hidden_dim
        self.num_heads = num_heads or 1
        self.head_size = hidden_dim//self.num_heads
        assert hidden_dim == self.head_size*self.num_heads, "hidden_dim must be divisible by num_heads"
        self.dropout = dropout
        self.bias = bias
        self.temperature = temperature
        self.use_se = use_se
        
        if self.positional_embedding == "dyn":
            self.dynpos = DynamicPositionBias(dim = hidden_dim//4, heads = num_heads, depth = 2)
        elif self.positional_embedding == "alibi":
            alibi_heads = num_heads // 2 + (num_heads % 2 == 1)
            self.alibi = AlibiPositionalBias(alibi_heads, self.num_heads)
            
        self.dropout_layer = nn.Dropout(dropout)
        self.weights = nn.Parameter(torch.empty(self.hidden_dim, 3 * self.hidden_dim)) # QKV
        self.out_w = nn.Parameter(torch.empty(self.hidden_dim, self.hidden_dim)) # QKV
        torch.nn.init.xavier_normal_(self.weights)
        torch.nn.init.xavier_normal_(self.out_w)

        if self.bias:
            self.out_bias = nn.Parameter(torch.empty(1, 1, self.hidden_dim)) # QKV
            self.in_bias = nn.Parameter(torch.empty(1, 1, 3*self.hidden_dim)) # QKV
            torch.nn.init.constant_(self.out_bias, 0.)
            torch.nn.init.constant_(self.in_bias, 0.)
        if not use_se:
            self.gamma = nn.Parameter(torch.ones(self.num_heads).view(1, -1, 1, 1))

    def forward(self, x, adj=None, mask=None, same=True, return_attn_weights=False):
        '''
            x: sequence feature: BxLxD
            adj: adj_matrix feature: Bx1xLxL
        '''
        b, l, h = x.shape
        x = x @ self.weights # b, l, 3*hidden
        if self.bias:
            x = x + self.in_bias
        Q, K, V = x.view(b, l, self.num_heads, -1).permute(0,2,1,3).chunk(3, dim=3) # b, a, l, head
        
        norm = self.head_size**0.5
        attention = (Q @ K.transpose(2,3)/self.temperature/norm)
        
        if self.positional_embedding == "dyn":
            i, j = map(lambda t: t.shape[-2], (Q, K))
            attn_bias = self.dynpos(i, j).unsqueeze(0)
            attention = attention + attn_bias
        elif self.positional_embedding == "alibi":
            i, j = map(lambda t: t.shape[-2], (Q, K))
            attn_bias = self.alibi(i, j).unsqueeze(0)
            attention = attention + attn_bias
        elif self.positional_embedding == "xpos":
            self.xpos = XPOS(self.head_size)

        if adj is not None:
            if not self.use_se:
                adj = self.gamma * adj
            attention = attention + adj

        attention = attention.softmax(dim=-1) # b, a, l, l, softmax won't change shape
        if mask is not None:
            ## (batch_size, seq_len)->(batch_size, seq_len, seq_len) -> (batch_size, num_heads, seq_len, seq_len)
            ## `&` op can be replaced by torch.bmm, more efficient for large computation. To Fix: "baddbmm_cuda" not implemented for 'Bool'
            mask2d = (mask.unsqueeze(1) & mask.unsqueeze(2)).unsqueeze(1).repeat(1, self.num_heads, 1, 1)
            # mask2d = mask.view(b,1,1,-1).repeat(1, self.num_heads, l, 1)  # wrong when broadcasting, reference for reproducing paper data
            attention = attention*mask2d
        '''
        ### Note: Different batch sizes result in slightly different outputs because of padding.
        ### There are many calculations that are influenced by padding: 
        1. another way for cal attetion:  attention = attention.masked_fill(~mask2d, float('-inf') => attention.softmax(dim=-1) =>  attention *= mask2d
        2. the out_bias may be masked: out_bias = out_bias*mask.unsqueeze(-1)
        3. the adj when apply conv: adj = adj*mask2d
        Since the differences are small, we ignore it.
        '''

        out = attention @ V  # b, a, l, head
        out = out.permute(0,2,1,3).flatten(2,3) # b, a, l, head -> b, l, (a, head) -> b, l, hidden
        if self.bias:
            out = out + self.out_bias
        if return_attn_weights:
            return out, attention
        else:
            return out           


class TransformerEncoderLayer(nn.Module):
    def __init__(self,
                 hidden_dim: int,
                 positional_embedding: str,
                 num_heads: int = None,
                 dropout: float = 0.10,
                 ffn_size: int = None,
                 activation: nn.Module = nn.GELU,
                 temperature: float = 1.,
                 use_se = False,
                ):
        super().__init__()
        if num_heads is None:
            num_heads = 1
        if ffn_size is None:
            ffn_size = hidden_dim*4
        self.in_norm = nn.LayerNorm(hidden_dim)
        self.mhsa = MultiHeadSelfAttention(hidden_dim=hidden_dim,
                                           num_heads=num_heads,
                                           positional_embedding=positional_embedding,
                                           dropout=dropout,
                                           bias=True,
                                           temperature=temperature,
                                           use_se=use_se,
                                          )
        self.dropout_layer = nn.Dropout(dropout)
        self.ffn = nn.Sequential(
            nn.LayerNorm(hidden_dim),
            nn.Linear(hidden_dim, ffn_size),
            activation(),
            nn.Dropout(dropout),
            nn.Linear(ffn_size, hidden_dim),
            nn.Dropout(dropout)
        )
        

    def forward(self, x, adj=None, mask = None, return_attn_weights = False):
        '''
            x: sequence feature: BxLXD
            adj: adj_matrix feature: Bx1xLxL
        '''
        x_in = x
        if return_attn_weights:
            x, attn_w = self.mhsa(self.in_norm(x), adj=adj, mask=mask, return_attn_weights = True)
        else:
            x = self.mhsa(self.in_norm(x), adj=adj, mask=mask, return_attn_weights = False)
        x = self.dropout_layer(x) + x_in
        x = self.ffn(x) + x

        if return_attn_weights:
            return x, attn_w
        else:
            return x
        

class AdjTransformerEncoder(nn.Module):
    def __init__(self,
                 positional_embedding: str,
                 dim: int  = 192,
                 head_size: int = 32,
                 dropout: float = 0.10,
                 dim_feedforward: int = 192 * 4,
                 activation: nn.Module = nn.GELU,
                 temperature: float = 1.,
                 num_layers: int = 12,
                 num_adj_convs: int =3,
                 ks: int = 3,
                 use_se = False,
                 conv_in_chan=0,
                ):
        super().__init__()
        num_heads, rest = divmod(dim, head_size)
        assert rest == 0
        self.num_heads = num_heads
        
        self.layers = nn.Sequential(
            *[TransformerEncoderLayer(hidden_dim=dim,
                                     num_heads=num_heads,
                                     positional_embedding=positional_embedding,
                                     dropout=dropout,
                                     ffn_size=dim_feedforward,
                                     activation=activation,
                                     temperature=temperature,
                                     use_se=use_se,
                                    ) 
             for i in range(num_layers)]
        )
        self.conv_in_chan = conv_in_chan
        if conv_in_chan>0:
            self.conv_layers = nn.ModuleList()
            for i in range(num_adj_convs):
                self.conv_layers.append(ResConv2dSimple(in_c=conv_in_chan if i == 0 else num_heads, out_c=num_heads, kernel_size=ks, use_se=use_se))
            
            
    def forward(self, x, adj, mask=None):
        '''
            x: sequence feature: BxLxD
            adj: adj_matrix feature: Bx1xLxL
        '''
        for ind, mod in enumerate(self.layers):
            if self.conv_in_chan>0:
                if ind < len(self.conv_layers):
                    adj = self.conv_layers[ind](adj)
            x = mod(x, adj=adj, mask=mask)
        return x

        
class BPfold_model(nn.Module):
    def __init__(self,  
                 positional_embedding: str,
                 not_slice: bool,
                 num_convs,
                 adj_ks=3,
                 dim=192, 
                 depth=12,
                 head_size=32,
                 use_se=False,
                 embed_filter=False,
                 use_BPP=True,
                 use_BPE=True,
                 dispart_outer_inner=True,
                 *args,
                 **kargs,
                 ):
        super().__init__()
        self.slice_tokens = not not_slice  # removing unnecessary padding tokens
        if num_convs is None:
            num_convs = depth
        # 4 nucleotides + 3 tokens: AUGC, BEGIN, END, EMPTY
        self.emb = nn.Embedding(4+3, dim)


        conv_in_chan = 0
        if use_BPP:
            conv_in_chan += 1
        if use_BPE:
            conv_in_chan += 1 + dispart_outer_inner

        self.transformer = AdjTransformerEncoder(
            num_layers=depth,
            num_adj_convs=num_convs,
            dim=dim,
            head_size=head_size,
            positional_embedding=positional_embedding,
            ks=adj_ks,
            use_se=use_se,
            conv_in_chan=conv_in_chan,
        )
        
        # self.proj_out = nn.Sequential(nn.Linear(dim, dim), nn.GELU(), nn.Linear(dim, 2))
        
        self.struct_embeds = nn.ModuleDict()
        
        self.use_BPP = use_BPP
        self.use_BPE = use_BPE
        self.embed_filter = embed_filter
        if self.embed_filter:
            self.filter_embedding = nn.Embedding(2, dim)
            
    def forward(self, x0):
        forward_mask = x0['forward_mask']
        x = x0['input'] # BxL
        mat = None
        if self.slice_tokens:
            forward_batch_Lmax = forward_mask.sum(-1).max()
            forward_mask = forward_mask[:,:forward_batch_Lmax]
            x = self.emb(x[:, :forward_batch_Lmax]) # x: B x forward_batch_Lmax x D
            if self.use_BPP:
                BPPM = x0['BPPM'][:, :, :forward_batch_Lmax, :forward_batch_Lmax] # Bx1xLxL -> B x 1 x batch_Lmax x batch_Lmax
                mat = torch.cat([mat, BPPM], dim=1) if mat is not None else BPPM
            if self.use_BPE:
                BPEM = x0['BPEM'][:, :, :forward_batch_Lmax, :forward_batch_Lmax] # Bx1xLxL -> B x 1 x batch_Lmax x batch_Lmax
                mat = torch.cat([mat, BPEM], dim=1) if mat is not None else BPEM
        if self.embed_filter:
            filter_embeded = self.filter_embedding(x0['is_good']).unsqueeze(1) # Bx1xE
            x = x + filter_embed
        
        x = self.transformer(x, mat, mask=forward_mask) # B x forward_batch_Lmax x D
        out = x @ x.transpose(-1, -2) # B x forward_batch_Lmax x forward_batch_Lmax
        return out


if __name__ == '__main__':
    arg_dic = {
               'adj_ks': 3, # int
               'depth': 12,
               'dim': 256,
               'head_size': 32,
               'not_slice': False,
               'num_convs': 3,
               'positional_embedding': 'dyn', # alibi
               'use_se': True,
              }
    Lmax = 128
    batch_size = 32
    x0 = {
          'forward_mask': torch.ones(batch_size, Lmax, dtype=bool),
          'input': torch.randint(0, 7, (batch_size, Lmax)),
          'BPEM': torch.randn((batch_size, 1, Lmax, Lmax)),
          'BPPM': torch.randn((batch_size, 1, Lmax, Lmax)),
         }
    model =  BPfold_model(**arg_dic)
    out = model(x0)
    print(model)
    print('out', out.shape)
    # for name, para in model.named_parameters():
    #     print(name)



import math

import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange


class DynamicPositionBias(nn.Module):
    def __init__(self, dim, *, heads, depth, log_distance=False, norm=False):
        super().__init__()
        assert depth >= 1, 'depth for dynamic position bias MLP must be greater or equal to 1'
        self.log_distance = log_distance

        self.mlp = nn.ModuleList([])

        self.mlp.append(nn.Sequential(
            nn.Linear(1, dim),
            nn.LayerNorm(dim) if norm else nn.Identity(),
            nn.SiLU()
        ))

        for _ in range(depth - 1):
            self.mlp.append(nn.Sequential(
                nn.Linear(dim, dim),
                nn.LayerNorm(dim) if norm else nn.Identity(),
                nn.SiLU()
            ))

        self.mlp.append(nn.Linear(dim, heads))

    @property
    def device(self):
        return next(self.parameters()).device

    def forward(self, i, j):
        assert i == j
        n, device = j, self.device

        # get the (n x n) matrix of distances
        seq_arange = torch.arange(n, device = device)
        context_arange = torch.arange(n, device = device)
        indices = rearrange(seq_arange, 'i -> i 1') - rearrange(context_arange, 'j -> 1 j')
        indices += (n - 1)

        # input to continuous positions MLP
        pos = torch.arange(-n + 1, n, device = device).float() # i==j, pos.shape == (2*i-1, 1)
        pos = rearrange(pos, '... -> ... 1')

        if self.log_distance:
            pos = torch.sign(pos) * torch.log(pos.abs() + 1)  # log of distance is sign(rel_pos) * log(abs(rel_pos) + 1)

        # after applying MLP to pos,  pos.shape == (2*i-1, num_heads), i==j
        for layer in self.mlp:
            pos = layer(pos)

        # get position biases        
        bias = pos[indices]
        bias = rearrange(bias, 'i j h -> h i j') #   bias.shape == (num_heads, i, i), i==j
        return bias

def pad_at_dim(t, pad, dim = -1, value = 0.):
    dims_from_right = (- dim - 1) if dim < 0 else (t.ndim - dim - 1)
    zeros = ((0, 0) * dims_from_right)
    return F.pad(t, (*zeros, *pad), value = value)

class AlibiPositionalBias(nn.Module):
    def __init__(self, heads, total_heads, **kwargs):
        super().__init__()
        self.heads = heads
        self.total_heads = total_heads

        slopes = torch.Tensor(self._get_slopes(heads))
        slopes = rearrange(slopes, 'h -> h 1 1')
        self.register_buffer('slopes', slopes, persistent = False)
        self.register_buffer('bias', None, persistent = False)
    
    def get_bias(self, i, j, device):
        i_arange = torch.arange(j - i, j, device = device)
        j_arange = torch.arange(j, device = device)
        bias = -torch.abs(rearrange(j_arange, 'j -> 1 1 j') - rearrange(i_arange, 'i -> 1 i 1'))
        return bias

    @staticmethod
    def _get_slopes(heads):
        def get_slopes_power_of_2(n):
            start = (2**(-2**-(math.log2(n)-3)))
            ratio = start
            return [start*ratio**i for i in range(n)]

        if math.log2(heads).is_integer():
            return get_slopes_power_of_2(heads)

        closest_power_of_2 = 2 ** math.floor(math.log2(heads))
        return get_slopes_power_of_2(closest_power_of_2) + get_slopes_power_of_2(2 * closest_power_of_2)[0::2][:heads-closest_power_of_2]

    @property
    def device(self):
        return next(self.buffers()).device

    def forward(self, i, j):
        h, device = self.total_heads, self.device

        if self.bias is not None and self.bias.shape[-1] >= j and self.bias.shape[-2] >= i:
            return self.bias[..., -i:, -j:]

        bias = self.get_bias(i, j, device)
        bias = bias * self.slopes

        num_heads_unalibied = h - bias.shape[0]
        bias = pad_at_dim(bias, (0, num_heads_unalibied), dim = 0)
        self.register_buffer('bias', bias, persistent = False)

        return self.bias


if __name__ == '__main__':
    hidden_dim = 16
    num_heads = 3
    dynamic_pe = DynamicPositionBias(dim=hidden_dim//4, heads=num_heads, depth=2)

    alibi_heads = num_heads // 2 + (num_heads % 2 == 1)
    alibi_pe = AlibiPositionalBias(alibi_heads, num_heads)
    i, j = 128, 128
    print('dyn bias shape', dynamic_pe(i, j).shape)
    print('alibi bias shape', alibi_pe(i, j).shape)



import os
import argparse

import numpy as np
import pandas as pd

from .util.base_pair_motif import BPM_energy
from .util.misc import get_file_name, str_localtime
from .util.RNA_kit import read_SS, connects2dbn, mut_seq, read_fasta, write_fasta 


def get_dbn(dest, src, gt_dir=None):
    src = os.path.abspath(src)
    len_src = len(src)
    sufs = ['bpseq', 'ct', 'fasta']
    sufs += [i.upper() for i in sufs]
    paths = [os.path.abspath(os.path.join(pre, f)) for pre, ds, fs in os.walk(src) for f in fs if any([f.endswith(suf) for suf in sufs])]
    with open(dest, 'w') as fp:
        ct = 0
        for src_path in paths:
            ct+=1
            name = get_file_name(src_path)
            seq, connects= read_SS(src_path)
            dbn= connects2dbn(connects)
            fp.write(f'>{name}\n')
            fp.write(f'{seq}\n')
            fp.write(f'{dbn}\n')
            if gt_dir:
                gt_path = os.path.join(gt_dir, src_path[len_src:].strip(os.path.sep))
                seq_gt, connects_gt = read_SS(gt_path)
                dbn_gt = connects2dbn(connects_gt)
                assert seq_gt.upper()==seq.upper(), f"{src_path}, {gt_path}\nseq1: {seq}\nseq2: {seq_gt}\n"
                fp.write(f'{dbn_gt} native\n')
    print(f'Result saved in {dest}.')


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input', type=str)
    parser.add_argument('-o', '--output', type=str, default='BPfold_kit_results')
    parser.add_argument('--print', action='store_true', help='print seq and SS')
    parser.add_argument('--get_fasta', action='store_true')
    parser.add_argument('--get_dbn', action='store_true', help='Meanwhile specify --gt_dir, saved in "./dbn_dataname.txt"')
    parser.add_argument('--get_matrix', type=str, choices=['energy', 'probability', ''], default='', help='input seq in shape of L, if energy: output nomralized outer BPM and inner BPM in shape of (2, L, L); if probability: output reference probability converted from energy in shape of (1, L, L).')
    parser.add_argument('--gt_dir', type=str)
    parser.add_argument('--name_col', type=str, default='target_id')
    parser.add_argument('--seq_col', type=str, default='sequence')
    parser.add_argument('--show_examples', action='store_true', help='Show examples and exit.')
    args = parser.parse_args()
    return args


def get_matrix(dest, name_seq_pairs, tag='energy'):
    assert tag in {'energy', 'probability'}, f'Unknown tag {tag}'
    dest = os.path.join(dest, f'{tag}_matrix')
    os.makedirs(dest, exist_ok=True)
    BPM_ene = BPM_energy()
    for name, seq in name_seq_pairs:
        seq = seq.upper().replace('T', 'U')
        m_seq = mut_seq(seq)
        if tag == 'energy':
            mat = BPM_ene.get_energy(m_seq, normalize_energy=True)
        else:
            mat = BPM_ene.get_probability(m_seq)
        np.save(os.path.join(dest, name+'.npy'), mat)


def main():
    args = parse_args()
    if args.show_examples:
        print('BPfold_kit --input example.bpseq --print')
        print('BPfold_kit --input data_dir --get_fasta')
        print('BPfold_kit --input data.csv --get_fasta --name_col target_id --seq_col sequence')
        print('BPfold_kit --input data_dir --get_dbn')
        print('BPfold_kit --input data_dir --get_dbn --gt_dir gt_dir')
        print('BPfold_kit --input seq --get_matrix energy')
        exit()
    os.makedirs(args.output, exist_ok=True)
    if os.path.isdir(args.input):
        save_name = os.path.basename(args.input)
    else:
        save_name = get_file_name(args.input)
    dest_path = None
    if args.print:
        seq, connects = read_SS(args.input)
        dbn = connects2dbn(connects)
        print(seq)
        print(dbn)
        dest_path = os.path.join(args.output, f'{save_name}.dbn')
        with open(dest_path, 'w') as fp:
            fp.write(f'>{save_name}\n{seq}\n{dbn}')
    if args.get_fasta:
        if os.path.isdir(args.input):
            name_seq_pairs = []
            for pre, ds, fs in os.walk(args.input):
                for f in fs:
                    seq, _ = read_SS(os.path.join(pre, f))
                    name = get_file_name(f)
                    name_seq_pairs.append((name, seq))
        elif args.input.endswith('.csv'):
            df = pd.read_csv(args.input)
            name_seq_pairs = zip(df[args.name_col], df[args.seq_col])
        else:
            raise Exception(args.input)
        dest_path = os.path.join(args.output, f'{save_name}.fasta')
        write_fasta(dest_path, name_seq_pairs)
    if args.get_dbn:
        dest_path = os.path.join(args.output, f'dbn_{save_name}.txt')
        get_dbn(dest=dest_path, src=args.input, gt_dir=args.gt_dir)
    if args.get_matrix:
        dest_path = args.output
        if os.path.exists(args.input): # fasta file
            get_matrix(args.output, read_fasta(args.input), args.get_matrix)
        else:
            time_str = str_localtime()
            get_matrix(args.output, [(f'seq_{time_str}', args.input)], args.get_matrix) # seq

    if dest_path is not None and os.path.exists(dest_path):
        print(f'Saving results in "{dest_path}"')


if __name__ == '__main__':
    main()



import os
import time
import random


def str_localtime(str_format="%Y%m%d_%Hh%Mm%Ss"):
    return time.strftime(str_format, time.localtime())


def read_dir(path, suffix_list=None):
    ret = []
    for pre, ds, fs in os.walk(path):
        for f in fs:
            if suffix_list is None or any([f.endswith(suf) for suf in suffix_list]):
                ret.append(os.path.join(pre, f))
    return ret


def get_file_name(path, flag='.', return_suf=False):
    file_name = os.path.basename(path)
    p = file_name.rfind(flag)
    suf = ''
    if p!=-1:
        suf = file_name[p:]
        file_name = file_name[:p]
    if return_suf:
        return file_name, suf
    else:
        return file_name


def timer(func):
    def ret_func(*args, **kargs):
        start = time.time()

        # execute func
        ret = func(*args, **kargs)

        end = time.time()
        res_time = int(end - start)
        units = ['d', 'h', 'm', 's']
        bases = [60*60*24, 60*60, 60, 1]

        nums = [0] * len(units)
        for i in range(len(units)):
            if res_time>bases[i]:
                nums[i] = res_time // bases[i]
                res_time = res_time%bases[i]
        time_str_lst = [f'{num}{unit}' for num, unit in zip(nums, units) if num!=0]
        if len(time_str_lst)>2:
            time_str_lst = time_str_lst[:2]
        time_str = ' '.join(time_str_lst)
        print(f'Time used: {time_str}')
        return ret
    return ret_func


def detect_encoding(path):
    import chardet
    with open(path, 'rb') as f:
        cont = f.read()
        result = chardet.detect(cont)
    return result['encoding']


def seed_everything(seed=42):
    import torch
    import numpy as np
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = True



import os
import yaml
import argparse

import torch
import numpy as np

from .util.yaml_config import read_yaml, write_yaml
from .util.misc import timer, get_file_name
from .util.RNA_kit import read_SS, dispart_nc_pairs, merge_connects, connects2mat, cal_metric


def get_index_data(gt_dir, testsets):
    index_path = None
    formats = ['bpseq', 'ct', 'dbn', 'BPSEQ', 'CT', 'DBN']
    for index_name in ['data_index.yaml']:
        cur_path = os.path.join(gt_dir, index_name)
        if os.path.exists(cur_path):
            index_path = cur_path
            break
    index_data = {}
    if index_path is None:
        for dataset in os.listdir(gt_dir):
            dataset_dir = os.path.join(gt_dir, dataset)
            if os.path.isdir(dataset_dir):
                if dataset.startswith('PDB_test'):
                    dataset = 'PDB_test'
                if testsets and dataset not in testsets:
                    continue
                if dataset not in index_data:
                    index_data[dataset] = []

                for pre, ds, fs in os.walk(dataset_dir):
                    for f in fs:
                        if any([f.endswith(fmt) for fmt in formats]):
                            index_data[dataset].append(os.path.join(pre, f))
            else:
                dataname = os.path.basename(gt_dir)
                index_data[dataname] = []
                for pre, ds, fs in os.walk(gt_dir):
                    for f in fs:
                        if any([f.endswith(fmt) for fmt in formats]):
                            index_data[dataname].append(os.path.join(pre, f))
                break
    else:
        for dic in read_yaml(index_path)['test']:
            dataset = dic['dataset']
            if dataset.startswith('PDB_test'):
                dataset = 'PDB_test'
            if testsets is not None and dataset not in testsets:
                continue
            if dataset not in index_data:
                index_data[dataset] = []
            index_data[dataset].append(os.path.join(gt_dir, dic['path']))
    return index_data


@timer
def evaluate(dest_path:str, pred_dir:str, gt_dir:str, read_pred=None, testsets=None, show_nc=False)->None:
    '''
    Cal metrics of predicted SS in pred_dir and gt_dir, then save results in dest_path (.yaml).

    Parameters
    ----------
    dest_path : str
        Path of yaml file to save the metric results
    pred_dir : str
        Path of pred SS files. Directory structure:
        -- pred_dir
            -- Dataset1
                -- file1.bpseq/.ct/.dbn/.out
                -- file2.bpseq/.ct/.dbn/.out
            -- Dataset2
    gt_dir : str
        Path of gt bpseq files. Directory structure is the same as pred_dir.
    '''
    SS_sufs = ['.bpseq', '.ct', '.dbn', '.out']
    SS_sufs += [i.upper() for i in SS_sufs]
    metric_dic = {}
    missing_pred = {}
    for dataset, gt_files in get_index_data(gt_dir, testsets).items():
        pred_data_dir = os.path.join(pred_dir, dataset)
        metric_dic[dataset] = {}
        for gt_path in gt_files:
            name = get_file_name(gt_path)
            pred_paths = [os.path.join(pred_pre, name+suf) for suf in SS_sufs for pred_pre in {pred_data_dir, pred_dir}]
            for pred_path in pred_paths:
                if os.path.exists(pred_path) and os.path.getsize(pred_path):
                    seq, pred_connects, gt_connects = get_seq_and_pred_gt_connects(pred_path, gt_path, read_pred)
                    canonical_pred, nc_pred = dispart_nc_pairs(seq, pred_connects)
                    canonical_gt, nc_gt = dispart_nc_pairs(seq, gt_connects)
                    metric_dic[dataset][name] = get_metric_dic(canonical_pred, canonical_gt)
                    if dataset.startswith('PDB_test'): # or any([i!=0 for i in nc_pred]):
                        # nc: non-canonical
                        nc_pred_path = os.path.join(os.path.dirname(pred_path), get_file_name(pred_path) + '_nc.bpseq')
                        if os.path.exists(nc_pred_path): 
                            _, nc_pred = read_SS(nc_pred_path)
                            pred_connects = merge_connects(canonical_pred, nc_pred)
                        pred_gt_dic = {'': {'pred': canonical_pred, 'gt': canonical_gt}, 
                                       '_nc': {'pred': nc_pred, 'gt': nc_gt},
                                       '_mix': {'pred': pred_connects, 'gt': gt_connects},
                                      }
                        types = ['', '_nc', '_mix'] if show_nc else ['']
                        for flag in types:
                            cur_dataset_str = dataset + flag
                            if cur_dataset_str not in metric_dic:
                                metric_dic[cur_dataset_str] = {}
                            metric_dic[cur_dataset_str][name] = get_metric_dic(pred_gt_dic[flag]['pred'], pred_gt_dic[flag]['gt'])
                    break
            else: # No pred results
                metric_dic[dataset][name] = {k: None for k in ['MCC', 'INF', 'F1', 'Precision', 'Recall', 'length']}
                missing_pred[name] = {'name': name, 'pred_dir': pred_dir, 'dataset': dataset}
    if missing_pred:
        print('missing pred', len(missing_pred))
        miss_path = os.path.join(os.path.dirname(dest_path), get_file_name(dest_path)+'_missing.yaml')
        write_yaml(miss_path, missing_pred)
    write_yaml(dest_path, metric_dic)


def get_metric_dic(pred_connects, gt_connects):
    m_dic = cal_metric(connects2mat(pred_connects), connects2mat(gt_connects))
    m_dic['length'] = len(gt_connects)
    return m_dic


def get_seq_and_pred_gt_connects(pred_path, gt_path, read_pred=None):
    '''
        pred_path, gt_path: .bpseq, .ct, .dbn, .out
        return: pred_connects, gt_connects
    '''
    if read_pred is None:
        read_pred = read_SS
    gt_bases, gt_connects = read_SS(gt_path)
    gt_seq = ''.join(gt_bases).upper()
    length = len(gt_seq)
    pred_bases, pred_connects = read_pred(pred_path)
    pred_seq = ''.join(pred_bases).upper()
    if len(gt_seq) != len(pred_seq):
        raise Exception(f'[Error] Lengthes of seqs dismatch: \n{gt_path}: {gt_seq}\n{pred_path}: {pred_seq}')
    if gt_seq != pred_seq:
        print(f'[Warning] Seq bases dismatch: {gt_path}, {pred_path}')
    return pred_seq, pred_connects, gt_connects



def summary(path, to_latex=True):
    metric_dic_all = None
    with open(path) as fp:
        metric_dic_all = yaml.load(fp.read(), Loader=yaml.FullLoader)
    metric_dic = {}
    metric_dic_gt600 = {}
    metric_dic_le600 = {}
    failed_dic = {}
    for dataset, dic in metric_dic_all.items():
        metric_dic[dataset] = {}
        metric_dic_le600[dataset] = {}
        metric_dic_gt600[dataset] = {}
        failed_dic[dataset] = 0
        for name, d_ori in dic.items():
            if any([v is None for v in d_ori.values()]):
                failed_dic[dataset] +=1
            d = {k: 0 if v is None else v for k,v in d_ori.items()}
            metric_dic[dataset][name] = d
            if d['length']<=600:
                metric_dic_le600[dataset][name] = d
            else:
                metric_dic_gt600[dataset][name] = d

    metric_names = ['INF', 'F1', 'Precision', 'Recall']
    pred_and_all = sorted([(dataset, len(dic)-failed_dic[dataset],len(dic)) for dataset, dic in metric_dic_all.items()])

    outputs = []
    outputs.append(f'[Summary] {path}')
    outputs.append(f' Pred/Total num: {pred_and_all}')
    for tag, cur_dic in [('len>600', metric_dic_gt600), ('len<=600', metric_dic_le600), ('all', metric_dic)]:
        title_row = format_row(['dataset', 'num', 'INF', 'F1', 'Precision', 'Recall'], [15, 5, 5, 5, 5, 5])
        length = (len(title_row) - len(tag))//2
        outputs.append('-'*length + tag + '-'*(len(title_row) - length - len(tag)))
        outputs.append(title_row)
        for dataset_name, dic in cur_dic.items():
            if len(dic)!=0:
                vals = [(metric, np.mean([d[metric] for d in dic.values()])) for metric in metric_names]
                items = [dataset_name, str(len(dic)), *[f'{v[-1]:.3f}' for v in vals]]
                outputs.append(format_row(items, [15, 5, 5, 5, 5, 5]))

    name = get_file_name(path)
    save_path = os.path.join(os.path.dirname(path), name+'_summary.txt')
    with open(save_path, 'w') as fp:
        for line in outputs:
            print(line)
            fp.write(line+'\n')


def format_row(items, len_adjust=None):
    if len_adjust is None:
        len_adjust = [0]*len(items)
    return ' & '.join([str(item).ljust(adj) for item, adj in zip(items, len_adjust)])+'\\\\'


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--pred_dir', type=str)
    parser.add_argument('--gt_dir', type=str, default='BPfold_data/test_data')
    parser.add_argument('--tag', type=str)
    parser.add_argument('--testsets', nargs='*')
    parser.add_argument('--show_nc', action='store_true')
    parser.add_argument('--summary', type=str, help='yaml file containing metric results.')
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    if args.pred_dir:
        if args.tag is None:
            segs = args.pred_dir.split(os.path.sep)[-3:]
            args.tag = 'eval_'+ '_'.join(segs)
        dest_path = args.tag+'.yaml'
        evaluate(dest_path, args.pred_dir, args.gt_dir, testsets=args.testsets, show_nc=args.show_nc)
        summary(dest_path)
    elif args.summary:
        summary(args.summary)


if __name__ == '__main__':
    main()



from functools import partial

import torch
import numpy as np
from fastai.vision.all import Metric

from ..util.postprocess import postprocess
from ..util.RNA_kit import _cal_metric_from_tp


def cal_loss(loss_func, pred, gt_dic, **args):
    '''
        pred: batch x batch_Lmax x batch_Lmax
        gt : batch x Lmax x Lmax
    '''
    forward_batch_Lmax = gt_dic['forward_mask'].sum(-1).max()
    batch_Lmax = forward_batch_Lmax - 2
    loss = loss_func(pred[:, 1:batch_Lmax+1, 1:batch_Lmax+1], gt_dic['gt'][:, 1:batch_Lmax+1, 1:batch_Lmax+1])
    return loss


def MSE_loss(pos_weight=300, **args):
    return partial(cal_loss, torch.nn.MSELoss(**args))


def BCE_loss(pos_weight=300, device=None, **args):
    pos_weight = torch.Tensor([pos_weight])
    if device is not None:
        pos_weight = pos_weight.to(device)
    loss_func = torch.nn.BCEWithLogitsLoss(pos_weight=pos_weight, **args)
    return partial(cal_loss, loss_func)


def cal_metric_torch(pred, gt):
    '''
        pred, gt: torch.Tensor
        return: MCC, INF, F1, precision, recall
    '''
    pred_p = torch.sign(pred).sum()
    gt_p = gt.sum()
    tp = torch.sign(pred*gt).sum()
    return _cal_metric_from_tp(torch.flatten(pred).shape[0], pred_p, gt_p, tp)


def cal_metric_batch(pred, gt, mask=None, seq_names=None, dataset_names=None):
    n = len(pred)
    if dataset_names is None:
        dataset_names = ['dataset' for i in range(n)]
    if seq_names is None:
        seq_names = [f'seq{i}' for i in range(n)]
    metric_dic = {dataset_name: {} for dataset_name in dataset_names}
    for i in range(n):
        dataset_name = dataset_names[i]
        seq_name = seq_names[i]
        cur_pred = pred[i] if mask is None else pred[i][mask[i]]
        cur_gt = gt[i] if mask is None else gt[i][mask[i]]
        m_dic = cal_metric_torch(cur_pred, cur_gt)
        metric_dic[dataset_name][seq_name] = {k: v.detach().cpu().numpy().item() for k, v in m_dic.items()}
    return metric_dic


class myMetric(Metric):
    def __init__(self, metric_name='F1', device=None): 
        '''
            metric_name: F1, MCC
        '''
        self.reset()
        self.metric_name = metric_name.upper()
        self.cal_func = cal_metric_batch
        
    def reset(self): 
        self.metrics = []
        
    def accumulate(self, learn):
        pred_batch = learn.pred
        data_dic = learn.y
        
        # prepare
        forward_batch_Lmax = data_dic['forward_mask'].sum(-1).max()
        batch_Lmax = forward_batch_Lmax-2
        pred_batch = pred_batch[:, 1:batch_Lmax+1, 1:batch_Lmax+1]
        mask = data_dic['mask'][:, 1:batch_Lmax+1, 1:batch_Lmax+1]
        seq_onehot = data_dic['seq_onehot'][:, 1:batch_Lmax+1, :]
        gt = data_dic['gt'][:, 1:batch_Lmax+1, 1:batch_Lmax+1]
        nc_map = data_dic['nc_map'][:, 1:batch_Lmax+1, 1:batch_Lmax+1]
        # postprocess
        ret_pred, _, ret_score, _ = postprocess(pred_batch, seq_onehot, nc_map, return_nc=False, return_score=False)
        metric_dic = self.cal_func(ret_pred, gt, mask)
        self.metrics += [d[self.metric_name] for dic in metric_dic.values() for d in dic.values()]

    @property
    def value(self):
        return np.mean(self.metrics)
    
    @property
    def name(self):
        return self.metric_name


def  get_loss(s):
    return {
            'bce': BCE_loss,
            'mse': MSE_loss,
           }[s.lower()]



import torch


class LenMatchBatchSampler(torch.utils.data.BatchSampler):
    def __iter__(self):
        buckets = {}
        yielded = 0

        for idx in self.sampler:
            s = self.sampler.data_source[idx]
            if isinstance(s, tuple): 
                L = s[0]["length"]
            else: 
                L = s["length"]
            L = max(1, L // 16) 
            if L not in buckets:  
                buckets[L] = []
            buckets[L].append(idx)
            
            if len(buckets[L]) == self.batch_size:
                batch = list(buckets[L])
                yield batch
                yielded += 1
                buckets[L] = []
                
        batch = []
        leftover = [idx for bucket in buckets.values() for idx in bucket]

        for idx in leftover:
            batch.append(idx)
            if len(batch) == self.batch_size:
                yielded += 1
                yield batch
                batch = []

        if len(batch) > 0 and not self.drop_last:
            yielded += 1
            yield batch


class DeviceMultiDataLoader:
    def __init__(self, dataloader_list, device='cuda', keywords=None):
        if not isinstance(dataloader_list, list):
            raise Exception('[Error] List of dataloaders expected.')
        self.device = device
        self.keywords = keywords
        self.dataloader_list = dataloader_list
        self.length = sum([len(l) for l in self.dataloader_list])
    
    def __len__(self):
        return self.length
    
    def __iter__(self):
        for dataloader in self.dataloader_list:
            for batch in dataloader:
                yield tuple({k: x[k].to(self.device) if self.keywords is None or k in self.keywords else x[k] for k in x} for x in batch)



import os
from itertools import product

import numpy as np

from .RNA_kit import mut_seq


SRC_DIR = os.path.dirname(os.path.realpath(__file__))
BPM_path = os.path.join(os.path.dirname(SRC_DIR), 'paras', 'key.energy')

def gen_all_BPMs(canonical_pairs=None, radius:int=3):
    '''
    Generate all outer and inner base pair motifs of six kinds of canonical pairs.
    inner hairpin base pair motif: BPM_iH,
    inner chain-break  base pair motif: BPM_iCB,
    outer chain-break base pair motif: BPM_oCB,

    Parameters
    ----------
    canonical_pairs: list
    list of canonical_pairs, default=['GC', 'CG', 'AU', 'UA', 'GU', 'UG']

    radius: int=3

    Returns
    -------
    yield: bpm
    '''
    assert radius>=2
    if canonical_pairs is None:
        canonical_pairs = ['GC', 'CG', 'AU', 'UA', 'GU', 'UG']
    for pair in canonical_pairs:
        yield from gen_base_pair_motif(pair, radius)


def gen_base_pair_motif(pair:str, radius:int=3):
    '''
    Generate  outer and inner base pair motifs of one input pair.
    inner hairpin base pair motif: BPM_iH,
    inner chain-break  base pair motif: BPM_iCB,
    outer chain-break base pair motif: BPM_oCB,

    Parameters
    ----------
    pair: str
    radius: int=3

    Returns
    -------
    yield: bpm
    '''

    assert radius>=2
    bases = 'AUGC'
    '''
    1 1
    1 4
    1 16
    1 64
    4 1
    4 4
    4 16
    4 64
    16 1
    16 4
    16 16
    16 64
    64 1
    64 4
    64 16
    3129
    '''
    # outer base pair motif: extend outward of seq, no more than r neighbors  # 3129
    for l_len, r_len in product(range(4), range(4)):
        if l_len == r_len ==3:
            continue
        l_seqs = None
        r_seqs = None
        if l_len==0:
            l_seqs = ['']
        else:
            l_seqs = product(*[bases for i in range(l_len)])
        if r_len==0:
            r_seqs = ['']
        else:
            r_seqs = product(*[bases for i in range(r_len)])
        for l_seq, r_seq in product(l_seqs, r_seqs):
            li = [pair[0], *l_seq, *r_seq, pair[1], '_', '0', '_', str(l_len+r_len+1), '-', str(l_len)] # TODO, cb+1?
            yield ''.join(li)

    # inner chain-break base pair motif: dis > 2*r,  # 4096
    d = 2*radius
    for seq in product(*[bases for i in range(2*radius)]):
        li = [pair[0], *seq[:radius], *seq[radius:], pair[1], '_', '0', '_', str(2*radius+1), '-', str(radius)] # TODO< cb+1?
        yield ''.join(li)

    ## inner hairpin base pair motif: dis <= 2*r,  # 5440 = 4^3 ... + 4^6
    for loop_len in range(3, 2*radius+1): # start with 2, since no sharp loop: |i-j|>=2
        for seq in product(*[bases for i in range(loop_len)]):
            li = [pair[0], *seq, pair[1], '_', '0', '_', str(loop_len+1)]
            yield ''.join(li)


def get_BPM_type(BPM_seq:str, radius:int=3)->str:
    if '-' in BPM_seq:
        if BPM_seq.endswith(f'{2*radius+1}-{radius}'):
            return 'BPM_iCB'
        else:
            return 'BPM_oCB'
    else:
        return 'BPM_iH'


def get_BPM_seqs(seq:str, i:int, j:int, r:int=3)->(str, str):
    '''
    Generate r-neighbor base pair motif for base pair (i, j) of seq
        ------i-----
                   |
        ------j-----

    Parameters
    ----------
    seq: str
        RNA sequence
    i: int
        index of base pair
    j: int
        index of base pair
    r: int
        r-neighbor

    Returns
    -------
    (bpm1, bpm2): (str, str)
        base pair motifs
    '''
    if i>j:
        i, j = j, i
    L = len(seq)
    # base pair motif 1: outer base pair motif
    left = max(i-r, 0)
    right = min(j+r, L-1)
    bpm_seq1 = seq[j: right+1] + seq[left: i+1]
    bpm1 = bpm_seq1 + f'_0_{len(bpm_seq1)-1}-{right-j}'

    # base pair motif 2: inner base pair motif
    bpm2 = None
    if j-i<=2*r: # hairpin loop
        bpm2 = seq[i:j+1] + f'_0_{j-i}'
    else:
        bpm2 = seq[i: i+r+1] + seq[j-r:j+1] + f'_0_{2*r+1}-{r}'
    return bpm1, bpm2


class BPM_energy:
    def __init__(self, path=BPM_path, radius=3):
        '''
            inner hairpin base pair motif: BPM_iH,
            inner chain-break  base pair motif: BPM_iCB,
            outer chain-break base pair motif: BPM_oCB,
        '''
        self.radiu = radius
        self.energy_table = {}
        self.norm_energy_table = {}
        # minE_H_CB_L_dic   ## min energy is minus
        minE_H_CB_L_dic = {}
        with open(path) as fp:
            for line in fp.readlines():
                line = line.strip('\n ')
                if line and not line.startswith('#'):
                    BPM, energy = line.split()
                    self.energy_table[BPM] = float(energy)
                    H_CB_L = BPM[BPM.find('_'):]
                    if H_CB_L in minE_H_CB_L_dic:
                        ## min energy is minus
                        minE_H_CB_L_dic[H_CB_L] = min(minE_H_CB_L_dic[H_CB_L], self.energy_table[BPM])
                    else:
                        minE_H_CB_L_dic[H_CB_L] = self.energy_table[BPM]
        for BPM in self.energy_table:
            H_CB_L = BPM[BPM.find('_'):]
            self.norm_energy_table[BPM] = self.energy_table[BPM]/minE_H_CB_L_dic[H_CB_L]
        #                       total negative
        # self.energy_table      76086 51184
        # self.norm_energy_table 76086 24902

    def get_energy_by_type(self, energy_table, BPM_seq:str, BPM_type:str='all'):
        cur_type = get_BPM_type(BPM_seq)
        if BPM_type == 'all' or cur_type==BPM_type:
            return energy_table[BPM_seq]
        else:
            return 0

    def get_energy(self, seq:str, radius:int=3, normalize_energy:bool=True, dispart_outer_inner=True, return_BPM:bool=False, BPM_type:str='all'):
        '''
        Generate energy map in shape of LxL according to input seq (L) and energy table

                norm  True          False
         dispart   
          True       [-1, 1]      same as (True, True)   2xLxL
          False    prob [0, 1]     ori energy            1xLxL

        Parameters
        ----------
        seq: str
            RNA sequence
        radius: int
            r-neighbor
        BPM_type: str
            all, BPM_iH, BPM_iCB, BPM_oCB

        Returns
        -------
        energy_map: np.ndarray
        '''
        assert BPM_type in ['all', 'BPM_iH', 'BPM_iCB', 'BPM_oCB'], f'[Error] Unknown BPM type: {BPM_type}'
        seq = mut_seq(seq)
        assert set(seq).issubset(set('AUGC')), f'Unknown base: seq="{seq}", should be A, U, G, C'
        canonical_pairs = {'AU', 'UA', 'GC', 'CG', 'GU', 'UG'}
        L= len(seq)
        mat = np.zeros((2, L, L)) if dispart_outer_inner else np.zeros((1, L, L))
        ret_BPM = {pair+'_0_1-0' for pair in canonical_pairs}
        for i in range(L):
            for j in range(L):
                pair = seq[i]+seq[j]
                if i==j or abs(j-i)<=3 or pair not in canonical_pairs:
                    if dispart_outer_inner: # TODO, non energy element should not be zero
                        pass
                    else:
                        if normalize_energy:
                            mat[:, i, j] = np.nan
                elif i<j:
                    bpm1, bpm2 = get_BPM_seqs(seq, i, j, r=radius)
                    if dispart_outer_inner:
                        mat[0][i][j] = self.get_energy_by_type(self.norm_energy_table, bpm1, BPM_type)
                        mat[1][i][j] = self.get_energy_by_type(self.norm_energy_table, bpm2, BPM_type)
                    else:
                        mat[0][i][j] =   self.get_energy_by_type(self.energy_table, bpm1, BPM_type)\
                                     + self.get_energy_by_type(self.energy_table, bpm2, BPM_type)\
                                     - self.get_energy_by_type(self.energy_table, pair+'_0_1-0', BPM_type)
                    ret_BPM.add(bpm1)
                    ret_BPM.add(bpm2)
                else:
                    mat[:, i, j] = mat[:, j, i]
        if not dispart_outer_inner and normalize_energy:
            # probability: E ~ = -lnp
            max_value = np.nanmax(np.abs(mat))
            if max_value>0 and not np.isnan(max_value):
                x = mat/max_value
                mat = np.exp(-x)/np.nansum(np.exp(-x)) # softmax, E ~= -lnp
            mat = np.nan_to_num(mat, nan=0)
        if return_BPM:
            return mat, ret_BPM
        else:
            return mat

    def get_probability(self, seq:str):
        return self.get_energy(seq, normalize_energy=True, dispart_outer_inner=False)


if __name__ == '__main__':
    BPM_ene = BPM_energy()
    seq_lst = ['AU', 'AUU', 'UG', 'AA', 'AUGAC', 'AUGCGUUCCAUTUGNa']
    for seq in seq_lst:
        print(seq)
        mat = BPM_ene.get_energy(seq)
        norm_mat = BPM_ene.get_energy(seq, normalize_energy=True)
        print(BPM_ene.get_probability(seq))
        for arr in mat:
            print(arr.tolist())
        print('norm mat 0')
        for arr in norm_mat[0]:
            print(arr.tolist())
        print('norm mat 1')
        for arr in norm_mat[1]:
            print(arr.tolist())

    type_dic = {}
    for BPM in BPM_ene.norm_energy_table:
        tp = get_BPM_type(BPM)
        if tp not in type_dic:
            type_dic[tp] = set()
        type_dic[tp].add(BPM)
    print({k:len(v) for k,v in type_dic.items()})
    print({k:len(v)//6 for k,v in type_dic.items()})
    for r in [2, 3, 4]:
        print(f'r={r}: {len(list(gen_all_BPMs(radius=r)))}')



import os
import yaml
import random
from itertools import product

import numpy as np
import torch
import torch.utils.data as data
from sklearn.model_selection import KFold

from ..util.misc import get_file_name, str_localtime
from ..util.RNA_kit import read_SS, connects2mat, mut_seq
from ..util.postprocess import get_base_index
from ..util.base_pair_motif import BPM_energy
from ..util.base_pair_probability import read_BPPM, gen_BPPM


CANONICAL_PAIRS = {'AU', 'UA', 'GC', 'CG', 'GU', 'UG'}


class RNAseq_data(data.Dataset):
    def __init__(self, 
                 data_dir,
                 index_name='data_index.yaml',
                 phase='train',
                 Lmax=600,
                 Lmin=0,
                 fold=0,
                 nfolds=4,
                 seed=42,
                 cache_dir='.cache_data', 
                 mask_only=False,
                 method='CDPfold',
                 trainall=False,
                 predict_files=None,
                 training_set=None,
                 test_set=None,
                 use_BPE=True, 
                 use_BPP=True, 
                 normalize_energy=False,
                 para_dir=None,
                 BPM_type='all',
                 verbose=True,
                 *args,
                 **kargs,
                 ):
        # Set all input args as attributes
        self.__dict__.update(locals())
        self.use_BPE = use_BPE
        self.use_BPP = use_BPP
        self.BPM_type = BPM_type
        self.phase = phase.lower()
        self.verbose = verbose
        if self.phase == 'predict':
            self.cache_dir = os.path.join(cache_dir, method)
        else:
            self.cache_dir = os.path.join(data_dir, cache_dir, method)
        self.method = method
        self.data_dir = data_dir
        self.mask_only = mask_only
        if use_BPP:
            os.makedirs(self.cache_dir, exist_ok=True)

        # data filter
        index_file = os.path.join(data_dir, index_name)
        if self.phase == 'predict':
            self.file_list = predict_files
            self.Lmax = max([f['length'] for f in self.file_list]) if predict_files else float('inf')
        else:
            if self.phase in {'train', 'validate'}:
                with open(index_file) as f:
                    index_dic = yaml.load(f.read(), Loader=yaml.FullLoader)
                    all_files = index_dic['train'][:]
                    if 'validate' in index_dic:
                        all_files += index_dic['validate']
                    if trainall:
                        all_files += index_dic['test']
                    # train with specific datasets
                    if self.phase == 'train' and training_set:
                        training_set = set(training_set)
                        if 'PDB_test' in training_set:
                            training_set.update({'PDB_test-TS1', 'PDB_test-TS2', 'PDB_test-TS3'})
                        self.myprint(f'training sets: {training_set}')
                        all_files = [f for f in all_files if f['dataset'] in training_set]

                    all_files.sort(key=lambda dic: dic['path'])

                    # Kfold
                    if nfolds <= 1: # no kfold
                        if self.phase == 'validate':
                            self.file_list = index_dic['test'][:]
                        else: # train
                            self.file_list = all_files[:]
                            random.shuffle(self.file_list)
                    else: # kfold training
                        split = list(KFold(n_splits=nfolds, random_state=seed, shuffle=True).split(range(len(all_files))))[fold][0 if self.phase == 'train' else 1]
                        self.file_list = [all_files[i] for i in split]
                self.Lmax = Lmax
                # limit length
                self.file_list = [f for f in self.file_list if Lmin<=f['length']<=self.Lmax]

            elif self.phase in ['test']:
                with open(index_file) as f:
                    index_dic = yaml.load(f.read(), Loader=yaml.FullLoader)
                    self.file_list = index_dic['test'][:]
                    if self.phase == 'test' and test_set:
                        test_set = set(test_set)
                        if 'PDB_test' in test_set:
                            test_set.update({'PDB_test-TS1', 'PDB_test-TS2', 'PDB_test-TS3'})
                        self.myprint(f'test sets: {test_set}')
                        self.file_list = [f for f in self.file_list if f['dataset'] in test_set]
                self.Lmax = max([f['length'] for f in self.file_list])
            else:
                raise NotImplementedError
            self.myprint(f'phase={self.phase}, num={len(self.file_list)}, nfolds={nfolds}: {index_file}')
            self.myprint(f'use_BPP={use_BPP}, use_BPE={use_BPE}')
            for dic in self.file_list:
                dic['path'] = os.path.join(self.data_dir, dic['path'])

        if para_dir is None:
            para_dir = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))), 'paras')
        if self.use_BPE:
            self.normalize_energy = normalize_energy
            self.BPM_ene = BPM_energy(path=os.path.join(para_dir, 'key.energy'))
        self.base_index = get_base_index() # for matrix embed
        self.num_base = len(self.base_index)
        self.index_base = {v: k for k, v in self.base_index.items()}
        self.token_index = {k: v for k, v in self.base_index.items()}
        self.token_index.update({tok: len(self.base_index)+i for i, tok in enumerate(['BOS', 'EOS', 'PAD'])}) # for sequence embed
        self.noncanonical = [self.index_base[i]+self.index_base[j] not in CANONICAL_PAIRS for i, j in product(range(self.num_base), range(self.num_base))]
        self.noncanonical_flag = np.array(self.noncanonical, dtype=bool)
        self.to_device_keywords = {'input', 'input_seqmat', 'mask', 'forward_mask', 'BPPM', 'BPEM', 'seq_onehot', 'nc_map',}
        if self.phase !='predict':
            self.to_device_keywords.add('gt')
            self.myprint(self.token_index)

    def myprint(self, *args):
        if self.verbose:
            print(*args)

    def __len__(self):
        return len(self.file_list)

    def prepare_data(self, name, seq, connects=None):
        ret = {}
        L = len(seq)
        ret['ori_seq'] = seq
        # ret['seq'] = ''.join([c if c in 'AUGCN' else 'N' for c in ret['ori_seq'].upper().replace('T', 'U')])
        ret['seq'] = mut_seq(seq)
        ret['length'] = L

        # mask: (Lmax+2)x(Lmax+2)
        mask = torch.zeros(self.Lmax + 2, self.Lmax+2, dtype=torch.bool)
        for row in range(1, L+1): # not including BOS and EOS
            mask[row, 1:L+1] = True
        # forward_mask: Lmax+2
        forward_mask = torch.zeros(self.Lmax + 2, dtype=torch.bool) # BOS, seq, EOS
        forward_mask[0:L+2] = True # including BOS and EOS
        ret['mask'] = mask
        ret['forward_mask'] = forward_mask

        if self.mask_only:
            return ret, None

        ## pad to uniform size=(Lmax+2) when batch-loading
        rside_pad = self.Lmax + 1 - L

        # seq_embed seq: Lmax+2  
        ret['input'] = self.seq_embed_sequence(ret['seq'])

        # seq_embed outer product 
        seqmat, seq_onehot = self.seq_embed_matrix(ret['seq'], return_onehot=True) # N**2xLxL
        seqmat_pad = np.pad(seqmat, ((0, 0), (1,rside_pad), (1, rside_pad)), constant_values=0)
        ret['input_seqmat'] = torch.FloatTensor(seqmat_pad) # NUM_BASE**2 x (Lmax+2) x (Lmax+2)
        seq_onehot_pad = np.pad(seq_onehot, ((1, rside_pad), (0, 0)), constant_values=0)
        ret['seq_onehot'] = torch.FloatTensor(seq_onehot_pad) # LxNUM_BASE

        # nc_map: noncanonical : (Lmax+2) x (Lmax+2)
        nc_map = seqmat[self.noncanonical_flag].sum(axis=0).astype(bool) # LxL
        nc_map_pad = np.pad(nc_map, ((1, rside_pad), (1, rside_pad)), constant_values=0)
        ret['nc_map'] = torch.FloatTensor(nc_map_pad)

        # BPPM: 1x(Lmax+2)x(Lmax+2)
        if self.use_BPP:
            bppm = self.load_BPPM(seq=seq, name=name, use_cache=(self.phase!='predict'))
            bppm_pad = np.pad(bppm, ((1, rside_pad), (1, rside_pad)), constant_values=0)
            ret['BPPM'] = torch.FloatTensor(bppm_pad).unsqueeze(0)
            # ret['BPPM'] = torch.log(ret['BPPM']+1e-5) # Note: Energy ~ log(p)
            # ret['BPPM'] = - ret['BPPM']
        # BPEM: 1x(Lmax+2)x(Lmax+2)
        if self.use_BPE:
            bpem = self.BPM_ene.get_energy(ret['seq'], normalize_energy=self.normalize_energy, BPM_type=self.BPM_type)
            bpem_pad = np.pad(bpem, ((0, 0), (1, rside_pad), (1, rside_pad)), constant_values=0)
            ret['BPEM'] = torch.FloatTensor(bpem_pad)

        y = {k: ret[k] for k in ['mask', 'forward_mask', 'nc_map', 'seq_onehot']}
        # gt, contact map: (Lmax+2)x(Lmax+2)
        if self.phase != 'predict':
            ### Important: N won't pair with other base
            for idx, conn in enumerate(connects):
                if conn!=0:
                    if ret['seq'][idx] == 'N':
                        connects[idx] = connects[conn-1] = 0

            gt = connects2mat(connects, strict=True)
            gt_pad = np.pad(gt, ((1, rside_pad), (1, rside_pad)), constant_values=0)
            ret['gt'] = torch.FloatTensor(gt_pad)
            y['gt'] = ret['gt']
        return ret, y


    def __getitem__(self, idx):
        info_dic = self.file_list[idx]
        dataset = info_dic['dataset'] if 'dataset' in info_dic else 'RNAseq'
        seq = name = connects = None
        
        if 'path' in info_dic:
            path = info_dic['path']
            name, suf = get_file_name(path, return_suf=True)
            if suf.lower() in {'.bpseq', '.ct', '.dbn'}:
                seq, connects = read_SS(path)
            else:
                with open(path) as fp:
                    fp.readline()
                    seq = fp.readline().strip(' \n')
        else:
            if 'seq' in info_dic:
                seq = info_dic['seq']
            else:
                raise Exception(f'[Error] seq or path needed: {info_dic}')
            name = info_dic['name'] if 'name' in info_dic else str_localtime()
        if connects is None and self.phase != 'predict':
            raise Exception(f'[Error] Invalid input: {info_dic} at {self.phase} stage, gt SS needed.')
        # # load_data
        # ret_data = {}
        # cache_path = os.path.join(self.cache_dir, name+'.pth')
        # if os.path.exists(cache_path):
        #     ret_data = torch.load(cache_path)
        # else:
        #     ret_data, y = self.prepare_data(name=name, seq=seq, connects=connects)
        #     if not self.mask_only:
        #         torch.save(ret_data, cache_path)
        ret_data, y = self.prepare_data(name=name, seq=seq, connects=connects)

        if self.mask_only:
            return {k: ret_data[k] for k in ['mask', 'forward_mask', 'length']}
        # update ret dic
        ret_data.update({'name': name, 'idx': idx, 'dataset': dataset})
        return ret_data, y

    def seq_embed_matrix(self, seq, return_onehot=False):
        ''' 
            seq: str, len=L, 'AUGC...'
            ret: tensor, NUM_BASE**2 x L x L, 0, 1 val
        '''
        L = len(seq)
        # seq onehot  L x NUM_BASE
        seq_onehot = np.zeros((L, self.num_base), dtype=float)
        for i in range(L): # should be consistent to function `postprocess` in ..util.postprocess
            seq_onehot[i][self.base_index[seq[i]]] = 1

        # seq embeding: NUM_BASE*NUM_BASE x L x L
        seq_embed = np.zeros((self.num_base**2, L, L)) # AUGC
        for n, (i, j) in enumerate(product(range(self.num_base), range(self.num_base))):
            seq_embed[n] = np.matmul(seq_onehot[:, i].reshape(-1, 1), seq_onehot[:, j].reshape(1, -1))
        if return_onehot:
            return seq_embed, seq_onehot
        else:
            return seq_embed

    def seq_embed_sequence(self, seq):
        ''' 
            seq: str, len=L, 'AUGC...'
            ret: ndarray, Lmax+2, 0-6 val, repr 'BOS AUGC N... end pad...'
        '''
        ret = [self.token_index['BOS']]
        ret.extend(self.token_index[s] for s in seq)
        ret.append(self.token_index['EOS'])
        
        # Lmax + 2, final length
        for i in range(self.Lmax - len(seq)): 
            ret.append(self.token_index['PAD'])
        ret = torch.Tensor(ret).int() # float? TODO
        return ret


    def load_BPPM(self, seq, name, use_cache=True):

        txt_path = os.path.join(self.cache_dir, name+'.txt')
        npy_path = os.path.join(self.cache_dir, name+'.npy')

        if use_cache and os.path.exists(npy_path):
            return np.load(npy_path, allow_pickle=True)
        else:
            if not os.path.exists(txt_path):
                try:
                    if self.phase == 'predict' and self.verbose:
                        self.myprint(f'[Info] Using "{self.method}" to generate BPPM, saving at "{txt_path}"')
                    gen_BPPM(txt_path, mut_seq(seq), name, self.method)
                except Exception as e:
                    if self.phase == 'predict' and self.verbose:
                        self.myprint(f'[Warning] {e}, using CDPfold instead')
                    gen_BPPM(txt_path, mut_seq(seq), name, 'CDPfold')
            BPPM = read_BPPM(txt_path, len(seq))
            np.save(npy_path, BPPM)
            return BPPM



import os
import math

import numpy as np

from .misc import get_file_name


def read_BPPM_ribonanza(path, L):
    ret = np.zeros((L, L))
    with open(path) as f:
        for line in f.readlines():
            line = line.strip('\n ')
            if line:
                row, col, val = line.split(' ')
                row, col, val = int(row)-1, int(col)-1, float(val)
                ret[row, col] = ret[col, row] = val
    return ret


def read_BPPM(path, L=None):
    '''
        Generated BPPM, same format of Contrafold's
    '''
    triples = []
    maxLen = 0
    with open(path) as f:
        for line in f.readlines():
            line = line.strip('\n ')
            if line:
                idx, base, *connects = line.split(' ')
                for connect_p in connects:
                    if connect_p:
                        connect, p = connect_p.split(':')
                        row, col, val = int(idx)-1, int(connect)-1, float(p)
                        triples.append((row, col, val))
                        maxLen = max(maxLen, col+1)
                maxLen = max(maxLen, int(idx))
    if L is None:
        L = maxLen
    ret = np.zeros((L, L), dtype=np.float32)
    for row, col, val in triples:
        ret[row, col] = ret[col, row] = val
    return ret


def write_BPPM(dest, seq, mat):
    '''
        write mat (BPPM of seq) to dest, same format of Contrafold's
        mat: BPPM
    '''
    L = len(seq)
    assert L!=0 and L == len(mat) == len(mat[0])
    with open(dest, 'w') as fp:
        for idx, base in enumerate(seq):
            BPP_str = []
            for j in range(idx+1, L):
                if mat[idx][j]>0:
                    BPP_str.append(f'{j+1}:{mat[idx][j]:.6f}')
            fp.write(f'{idx+1} {base} {" ".join(BPP_str)}\n')


def CDP_BPPM(seq):
    def gaussian(x):
        return math.exp(-0.5*(x*x))
    def get_score(baseA, baseB):
        if {baseA, baseB} == {'A', 'U'}:
            return 2
        elif {baseA, baseB} == {'G', 'C'}:
            return 3
        elif {baseA, baseB} == {'G', 'U'}:
            return 0.8
        else:
            return 0
    L = len(seq)
    seq = seq.upper().replace('T', 'U')
    mat = np.zeros([L, L])
    for i in range(L):
        for j in range(L):
            for add in range(30):
                if i - add >= 0 and j + add <L:
                    score = get_score(seq[i-add], seq[j+add])
                    if score == 0:
                        break
                    else:
                        mat[i,j] = score * gaussian(add)
                else:
                    break
            if mat[i,j] > 0:
                for add in range(1, 30):
                    if i + add < L and j - add >= 0:
                        score = get_score(seq[i+add], seq[j-add])
                        if score == 0:
                            break
                        else:
                            mat[i,j] = score * gaussian(add)
                    else:
                        break
    return mat


def read_BPP_from_ps(ps_path, threshold=0.001):
    beginSeq = False
    beginBPP = False
    seq_lines = []
    triples = []
    with open(ps_path) as fp:
        for line in fp.readlines():
            if beginSeq:
                if line.startswith(') } def'):
                    beginSeq = False
                else:
                    seq_lines.append(line.strip(' \n\r\\'))
            else:
                if line.startswith('/sequence { ('):
                    beginSeq = True
            if beginBPP:
                if line.startswith('showpage'):
                    beginBPP = False
                else:
                    i, j, v, _ = line.split(' ')
                    i, j, v = int(i)-1, int(j)-1, float(v)
                    if v>=threshold:
                        triples.append((i, j, v))
            else:
                if line.startswith('%start of base pair prob'):
                    beginBPP = True
    seq = ''.join(seq_lines)
    L = len(seq)
    mat = np.zeros((L, L), dtype=float)
    for i, j, v in triples:
        mat[i][j] = mat[j][i] = v
    return seq, mat


def gen_BPPM(dest_path:str, mutated_seq:str, name:str, method:str='CDPfold', posteriors:float=0.001, prefix_dir:str='./BPPM_generator')->None:
    '''
    Generate BPPM of seq_path, write to dest_path

    Parameters
    ----------
    dest_path: str
        dest path: *.txt
    mutated_seq: str
        only contains AUGC
    name: str
    method: str
        'EternaFold', 'CDPfold', 'Contrafold', 'ViennaRNA'
    '''
    log_path = os.path.join(prefix_dir, '.gen_BPPM.log')
    # write seq and ss
    mut_seq_path = os.path.join(os.path.dirname(dest_path), name+'.fasta')
    write_fasta(mut_seq_path, [(name, mutated_seq)])

    method_low = method.lower()
    if method_low in {'eternafold', 'contrafold'}:
        paras = f'--params {prefix_dir}/EternaFold/parameters/EternaFoldParams.v1' if method_low == 'eternafold' else ''
        cmd_str = f'{prefix_dir}/EternaFold/src/contrafold predict "{mut_seq_path}" {paras} --posteriors {posteriors} "{dest_path}" >> {log_path} 2>&1'
        ret_code = os.system(cmd_str)
        if ret_code:
            raise Exception(f'Error encountered when using {method}')
    elif method_low == 'cdpfold':
        mat = CDP_BPPM(mutated_seq)
        write_BPPM(dest_path, mutated_seq, mat)
    elif method_low == 'viennarna':
        prefix = os.path.join(prefix_dir, method)
        cmd_str = f'{prefix}/bin/RNAfold -p --noPS "{mut_seq_path}" >> {log_path} 2>&1'
        ret_code = os.system(cmd_str)
        if ret_code:
            raise Exception(f'Error encountered when using {method}')
        ps_path = get_file_name(dest_seq_name)+'_dp.ps'
        cur_seq, mat = read_BPP_from_ps(ps_path)
        write_BPPM(dest_path, cur_seq, mat)
    else:
        raise Exception(f'Unknown method {method}')



import re

import numpy as np

from .misc import get_file_name
from .base_pair_probability import CDP_BPPM


def read_fasta(path):
    '''
    Read iterator of all sequences from a fasta file.

    Parameters
    ----------
    path: str
        Path of fasta file.

    Returns
    -------
    yield: (str, str)
        (seq_name, seq)
    '''
    seq_name = None
    with open(path) as fp:
        for line in fp.readlines():
            line = line.strip(' \n\r\t')
            if line.startswith('#') or line=='':
                continue
            elif line.startswith('>'):
                seq_name = line[1:]
            else:
                yield seq_name, line


def write_fasta(path:str, name_seq_pairs:[(str, str)])->None:
    with open(path, 'w') as fp:
        for name, seq in name_seq_pairs:
            fp.write(f'>{name}\n{seq}\n')


def read_SS(path:str, return_index:bool=False):
    '''
    Read secondary structure from bpseq/ct/dbn file.

    Parameters
    ----------
    path: str
        path of ss file, endswith '.bpseq', '.ct', '.dbn'
    return_index : bool
        indicate whether this function returns indexes of all bases

    Returns
    -------
    seq: str, length L
        Containing RNA bases: AUGC.
    connects: [int], length L
        The i-th base connects to `connects[i-1]`-th base, 1-indexed, 0 for no connection.
    index: [int], length L
        list of base indexes, if valid, = list(range(1, 1+L))
    '''
    low_path = path.lower()
    if low_path.endswith('.bpseq'):
        return read_bpseq(path, return_index)
    elif low_path.endswith('.ct'):
        return read_ct(path, return_index)
    elif low_path.endswith('.dbn'):
        return read_dbn(path, return_index)
    else:
        raise Exception(f'[Error] Unkown file type: {path}')


def read_dbn(path:str, return_index:bool=False):
    seq = dbn = ''
    with open(path) as fp:
        for line in fp.readlines():
            line = line.strip('\r\n ')
            if not line or line.startswith('>') or line.startswith('#'):
                continue
            elif line[0].isalpha():
                if dbn:
                    break
                seq += line
            else:
                dbn += line
    connects = dbn2connects(dbn)
    if return_index:
        return seq, connects, list(range(1, 1+len(connects)))
    else:
        return seq, connects


def read_bpseq(path:str, return_index:bool=False):
    '''
    Read bpseq file.

    Parameters
    ----------
    path: str
        Path of bpseq file.
    return_index: bool
        indicate whether this function returns indexes of all bases

    Returns
    -------
    seq: str, length L
        Containing RNA bases: AUGC.
    connects: [int], length L
        The i-th base connects to `connects[i-1]`-th base, 1-indexed, 0 for no connection.
    index: [int], length L
        list of base indexes, if valid, = list(range(1, 1+L))
    '''
    bases = []
    connects = []
    indexes = []
    with open(path) as f:
        for line in f.readlines():
            if not line.startswith('#'):
                idx, base, conn = [item for item in line.strip('\n\t\r ').split() if item]
                bases.append(base)
                connects.append(conn)
                indexes.append(int(idx))
    connects = [int(i) for i in connects]
    if return_index:
        return ''.join(bases), connects, indexes
    else:
        return ''.join(bases), connects


def read_ct(path:str, return_index:bool=False):
    '''
    Read ct file.

    Parameters
    ----------
    path: str
        Path of ct file.
    return_index: bool
        indicate whether this function returns indexes of all bases

    Returns
    -------
    seq: str, length L
        Containing RNA bases: AUGC.
    connects: [int], length L
        The i-th base connects to `connects[i-1]`-th base, 1-indexed, 0 for no connection.
    index: [int], length L
        list of base indexes, if valid, = list(range(1, 1+L))
    '''
    bases = []
    connects = []
    indexes = []
    last_idx = 0
    pattern = r'(?P<idx>\d+)[\t ]+(?P<base>.)[\t ]+(?P<pre>\d+)[\t ]+(?P<nxt>\d+)[\t ]+(?P<conn>-?\d+)[\t ]+(?P<cur>\d+)'
    with open(path) as fp:
        for idx, base, pre, nxt, conn, cur in re.findall(pattern, fp.read()):
            # print(idx, base, pre, nxt, conn, cur)
            try:
                assert int(nxt)==0 or (int(cur) == int(nxt)-1 == int(pre)+1)
            except Exception as e:
                print(e)
                continue
            idx, conn = int(idx), int(conn)
            if conn < 0:
                conn = 0
            if idx!=last_idx+1:
                raise Exception(f'[Error] Inconsistent bases (line={last_idx}): {path}')
            last_idx = idx
            bases.append(base)
            connects.append(conn)
            indexes.append(idx)
    if return_index:
        return ''.join(bases), connects, indexes
    else:
        return ''.join(bases), connects


def write_SS(path:str, seq:str, connects:[int])->None:
    '''
    Write secondary structure to bpseq/ct/dbn.

    Parameters
    ----------
    path: str
        Dest path.
    seq: str, length L
        Containing RNA bases: AUGC.
    connects: [int], length L
        The i-th base connects to `connects[i-1]`-th base, 1-indexed, 0 for no connection.
    '''
    low_path = path.lower()
    if low_path.endswith('bpseq'):
        write_bpseq(path, seq, connects)
    elif low_path.endswith('ct'):
        write_ct(path, seq, connects)
    elif low_path.endswith('dbn'):
        write_dbn(path, seq, connects)
    else:
        raise Exception(f'Unkown file type: {path}, only .ct, .bpseq, .dbn are allowed.')


def write_dbn(path:str, seq:str, connects:[int])->None:
    '''
    Write secondary structure to dbn file.

    Parameters
    ----------
    path: str
        Dest path.
    seq: str, length L
        Containing RNA bases: AUGC.
    connects: [int], length L
        The i-th base connects to `connects[i-1]`-th base, 1-indexed, 0 for no connection.
    '''
    with open(path, 'w') as f:
        dbn = connects2dbn(connects)
        f.write(f'{seq}\n{dbn}\n')


def write_ct(path:str, seq:str, connects:[int])->None:
    '''
    Write secondary structure to ct file.

    Parameters
    ----------
    path: str
        Dest path.
    seq: str, length L
        Containing RNA bases: AUGC.
    connects: [int], length L
        The i-th base connects to `connects[i-1]`-th base, 1-indexed, 0 for no connection.
    '''
    with open(path, 'w') as f:
        for i, (base, connect) in enumerate(zip(seq, connects)):
            f.write(f'{i+1} {base.upper()} {i} {i+2} {connect} {i+1}\n')


def write_bpseq(path:str, seq:str, connects:[int], comments=None)->None:
    '''
    Write secondary structure to bpseq file.

    Parameters
    ----------
    path: str
        Dest path.
    seq: str, length L
        Containing RNA bases: AUGC.
    connects: [int], length L
        The i-th base connects to `connects[i-1]`-th base, 1-indexed, 0 for no connection.
    '''
    with open(path, 'w') as f:
        if comments is not None:
            f.write(f'# {comm}\n')
        for i, (base, connect) in enumerate(zip(seq, connects)):
            f.write(f'{i+1} {base.upper()} {connect}\n')


def dispart_nc_pairs(seq:str, connects:[int], extra_nc_positions=None)->[int]:
    '''
    Dispart canonical and noncanonical pairs in connects.

    Parameters
    ----------
    seq: str
        RNA base seq
    connects: [int], length L
        The i-th base connects to `connects[i-1]`-th base, 1-indexed, 0 for no connection.

    Returns
    -------
    connects: [int], length L
    nc_connects: [int], length L
    '''
    L = len(seq)
    assert L == len(connects) != 0
    seq = seq.upper()
    canonical_pairs = {'AU', 'UA', 'GC', 'CG', 'GU', 'UG'}
    conns = [0] * L
    nc_conns = [0] * L
    if extra_nc_positions is None:
        extra_nc_positions = [False] * L
    for idx, (base, conn) in enumerate(zip(seq, connects)):
        if conn!=0 and idx<conn-1:
            pair = base + seq[conn-1]
            if pair in canonical_pairs:
                conns[idx] = conn
                conns[conn-1] = idx+1
            else:
                nc_conns[idx] = conn
                nc_conns[conn-1] = idx+1
            # extra nc positions
            if extra_nc_positions[idx] or extra_nc_positions[conn-1]:
                nc_conns[idx] = conn
                nc_conns[conn-1] = idx+1
    return conns, nc_conns


def merge_connects(connects:[int], nc_connects:[int])->[int]:
    '''
    Merge canonical and noncanonical pairs in connects.

    Parameters
    ----------
    connects: [int], length L
        The i-th base connects to `connects[i-1]`-th base, 1-indexed, 0 for no connection.

    Returns
    -------
    connects: [int], length L
    '''
    n = len(connects)
    n1 = len(nc_connects)
    assert n==n1, f'length mismatch: {n}!={n1}'
    ret_connects = connects[:]
    dic = {i+1: conn for i, conn in enumerate(connects)}
    for i, conn in enumerate(nc_connects):
        if conn!=0 and i<conn:
            if connects[i]==0 and connects[conn-1]==0:
                ret_connects[i] = conn
                ret_connects[conn-1] = i+1
    return ret_connects


def remove_lone_pairs(connects:[int], loop_len:int=3)->[int]:
    '''
    Remove lone pairs in connects

    Parameters
    ----------
    connects: [int], length L
        The i-th base connects to `connects[i-1]`-th base, 1-indexed, 0 for no connection.
    loop_len: int
        interval of lone pair and other pair

    Returns
    -------
    connects: [int], length L
        The i-th base connects to `connects[i-1]`-th base, 1-indexed, 0 for no connection.
    '''
    L = len(connects)
    ret_conn = [0] * L
    for idx, conn in enumerate(connects):
        if conn!=0 and idx<conn-1:
            isLone = True
            for center, conj in [(idx, conn-1), (conn-1, idx)]:
                for dire in [-1, 1]:
                    neighbor = center + dire
                    if 0 <= neighbor < L:
                        if connects[neighbor]-1 in [conj-i*dire for i in range(1, loop_len+1)]:
                            isLone = False
                            break
                if not isLone:
                    break
            if not isLone:
                ret_conn[idx] = conn
                ret_conn[conn-1] = idx+1
    return ret_conn


def dbn2connects(dbn:str)->[int]:
    '''
    Convert dbn to connects.

    Parameters
    ----------
    dbn: str
        Dot-bracket notation of secondary structure, including '.()[]{}', with valid brackets.

    Returns
    -------
    connects: [int], length L
        The i-th base connects to `connects[i-1]`-th base, 1-indexed, 0 for no connection.
    '''
    alphabet = ''.join([chr(ord('A')+i) for i in range(26)])
    alphabet_low = alphabet.lower()
    syms = ('([{<' + alphabet)[::-1]
    syms_conj = (')]}>' + alphabet_low)[::-1]
    left2right = {p: q for p, q in zip(syms, syms_conj)}
    right2left = {p: q for p, q in zip(syms_conj, syms)}

    pair_dic = {}
    stack_dic = {char: [] for char in left2right}
    for i, char in enumerate(dbn):
        idx = i+1
        if char=='.':
            pair_dic[idx] = 0
        elif char in left2right:
            stack_dic[char].append((idx, char))
        elif char in right2left:
            cur_stack = stack_dic[right2left[char]]
            if len(cur_stack)==0:
                raise Exception(f'[Error] Invalid brackets: {dbn}')
            p, ch = cur_stack.pop()
            pair_dic[p] = idx
            pair_dic[idx] = p
        else:
            raise Exception(f'[Error] Unknown DBN representation: dbn[{i}]={char}: {dbn}')
    if any(stack for k, stack in stack_dic.items()):
        raise Exception(f'[Error] Brackets dismatch: {dbn}')
    connects = [pair_dic[i] for i in range(1, 1+ len(dbn))]
    return connects


def connects2dbn(connects:[int])->str:
    '''
    Convert connects to dbn. [Warning] Can't deal with pseudo knot.

    Parameters
    ----------
    connects: [int], length L
        The i-th base connects to `connects[i-1]`-th base, 1-indexed, 0 for no connection.

    Returns
    -------
    dbn: str
        Dot-bracket notation of secondary structure, including '.()[]{}', with valid brackets.
    '''
    alphabet = ''.join([chr(ord('A')+i) for i in range(26)])
    alphabet_low = alphabet.lower()
    syms = ('([{<' + alphabet)[::-1]
    syms_conj = (')]}>' + alphabet_low)[::-1]
    syms_set = set(syms)
    syms_conj_set = set(syms_conj)
    ret = ['.']*len(connects)
    for i, conn in enumerate(connects):
        pi = conn-1
        if pi != -1 and pi>=i:
            counts = [0] * len(syms)
            for j in range(i+1, pi):
                sym = ret[j]
                if sym in syms_set:
                    ct_idx = syms.index(sym)
                    counts[ct_idx] +=1
                elif sym in syms_conj_set:
                    ct_idx = syms_conj.index(sym)
                    counts[ct_idx] -=1
            for idx, ct in enumerate(counts):
                if ct==0:
                    ret[i] = syms[idx]
                    ret[pi] = syms_conj[idx]
    return ''.join(ret)


def mat2connects(mat, loop_min_len=2, greedy_max_matching=True)->[int]:
    '''
    Convert contact map to connects, enforce single pairing.

    Parameters
    ----------
    mat: numpy.ndarray
        LxL matrix where the mat[i,j]=score, the biger, the more probable to pair.

    Returns
    -------
    connects: [int], length L
        The i-th base connects to `connects[i-1]`-th base, 1-indexed, 0 for no connection.
    '''
    if greedy_max_matching:
        # enforce single pairing
        mat = greedy_max_weight_matching(mat)
    connects = mat.argmax(axis=1)
    connects[connects!=0] +=1
    if connects[0]!=0:
        connects[connects[0]-1] = 1
    return connects.tolist()


def greedy_max_weight_matching(M):
    '''
    Greedy max weight matching, fast, but may not be optimal.
    '''
    n = len(M)
    M_new = np.zeros((n, n), dtype=float)
    used = set()
    
    # Flatten and sort all possible pairs by score (descending)
    pairs = []
    for i in range(n):
        for j in range(i + 1, n):
            if M[i][j] > 0:
                pairs.append((M[i][j], i, j))
    pairs.sort(reverse=True, key=lambda x: x[0])
    
    # Assign pairs greedily
    for score, i, j in pairs:
        if i not in used and j not in used:
            M_new[i][j] = score
            M_new[j][i] = score
            used.add(i)
            used.add(j)
    return M_new


def nx_max_weight_matching(M):
    '''
    Use pacakge networkx, may be slow when length > 1000.
    '''
    import networkx as nx
    n = len(M)
    G = nx.Graph()
    
    # Add edges with weights (only upper triangle to avoid duplication)
    for i in range(n):
        for j in range(i + 1, n):
            if M[i][j] > 0:
                G.add_edge(i, j, weight=M[i][j])
    
    # Compute maximum weight matching
    matching = nx.max_weight_matching(G, maxcardinality=False)
    
    # Build new contact matrix
    M_new = np.zeros((n, n), dtype=float)
    for i, j in matching:
        M_new[i][j] = M[i][j]
        M_new[j][i] = M[j][i]
    return M_new


def connects2mat(connects:[int], strict=True):
    '''
    Convert connects to contact map.

    Parameters
    ----------
    connects: [int], length L
        The i-th base connects to `connects[i-1]`-th base, 1-indexed, 0 for no connection.
    strict: bool
        if not strict, ignore invalid connects (that exceeds L)

    Returns
    -------
    mat: numpy.ndarray
        LxL matrix where the mat[i,j]=1 represents paired bases, otherwise 0
    '''
    L = len(connects)
    ret = np.zeros((L, L))
    for num, conn in zip(range(1, L+1), connects):
        if conn!=0:
            if (conn-1>=L or num-1 >= L) and not strict:
                continue
            ret[num-1][conn-1] = ret[conn-1][num-1] = 1
    return ret


def read_react(path):
    '''
    Read reactivities from path. It contains white-space-delimited columns. The first column is the nucleotide, the second column is the chemical reactivity. NA reactivity data are denoted as number less than -100, 'NA', or 'nan'.

    Parameters
    ----------
    path: str
        path of reactivity file

    Returns
    -------
    reacts: [float]
    '''
    bases = []
    reacts = []
    with open(path) as fp:
        for line in fp.readlines():
            line = line.strip('\n\t\r ')
            if line:
                if line.startswith('#'):
                    continue
                elif line[0].isdigit():
                    parts = [part for part in line.replace('\t', ' ').split() if part]
                    assert len(parts) in {2,3}, f'[Error] when parsing line {line} from {path}.'
                    idx = int(parts[0])
                    base = react = None
                    if len(parts)==2:
                        react = parts[1]
                    else:
                        base = parts[1]
                        assert base in 'AUGCNT'
                        react = parts[2]
                    bases.append(base)
                    if len(reacts)+1!=idx:
                        reacts += [float('nan') for i in range(idx-1-len(reacts))]
                    if react.lower().startswith('n'):
                        reacts.append(float('nan'))
                    else:
                        reacts.append(float(react))
                else:
                    raise Exception('[Error] when reading react from {path}.')
    return reacts


def write_react(path, reacts, seq=None, delimiter=' '):
    '''
    Write reactivities of seq to path. It contains white-space-delimited columns. The first column is the nucleotide, the second column is the chemical reactivity. NA reactivity data are denoted as number less than -100, 'NA', or 'nan'.

    Parameters
    ----------
    path: str
        path of reactivity file
    reacts: [float]
        reactivities
    seq: str
        AUGCNT
    '''
    if seq is not None:
        assert len(seq) == len(reacts), f'length mismatch, seq={len(seq)}, reacts={len(reacts)}'
    with open(path, 'w') as fp:
        for idx in range(len(reacts)):
            if seq is None:
                fp.write(f'{idx+1}{delimiter}{reacts[idx]}\n')
            else:
                fp.write(f'{idx+1}{delimiter}{seq[idx]}{delimiter}{reacts[idx]}\n')


def valid_ss(seq:str, connects:[int], indexes:[int]=None)->bool:
    return len(seq)>10 and len(seq) == len(connects) and min(connects)>=0 and max(connects)<=len(seq) and (indexes is None or indexes==list(range(1, 1+len(seq))))


def is_valid_bracket(s, ignore_unknown=False):
    chars = set('.()[]{}')
    left2right = {p: q for p, q in ['()', '[]', '{}']}
    right2left = {q: p for p, q in ['()', '[]', '{}']}
    count_dic = {ch: 0 for ch in chars}

    for i, char in enumerate(s):
        count_dic[char] += 1
        if char in '.([{':
            pass
        elif char in ')]}':
            if count_dic[char] > count_dic[right2left[char]]:
                return False
        else:
            if not ignore_unknown:
                raise Exception(f'[Error] Unknown brackets repr: {s}')
    return all(count_dic[char]==count_dic[left2right[char]] for char in '([{')


def mut_seq(seq:str, connects=None)->str:
    '''
    Mutate unknown chars to conj/U.

    Parameters
    ----------
    seq: str, length L
        Containing RNA bases AUGC or other unknown chars.
    connects: [int] or None, length L
        The i-th base connects to `connects[i-1]`-th base, 1-indexed, 0 for no connection.

    Returns
    -------
    new_seq: str, length L
        Containing RNA bases AUGC only.
    '''
    seq = seq.upper().replace('T', 'U')
    if set(seq).issubset(set('AUGC')):
        return seq
    new_seq = []
    chars = {'A', 'U', 'G', 'C'}
    conj = {'A': 'U', 'U': 'A', 'G': 'C', 'C': 'G'}
    if connects is None:
        connects = mat2connects(CDP_BPPM(seq))
    for i in range(len(seq)):
        if seq[i] in chars:
            new_seq.append(seq[i])
        else:
            conn = connects[i]
            if conn!=0 and seq[conn-1] in chars:
                new_seq.append(conj[seq[conn-1]])
            else:
                new_seq.append('U')
    return ''.join(new_seq)


def mat2scores(mat, connects, mode)->[float]:
    '''
    NOTICE! Deprecated, discarded!

    Convert contact map to prob scores.

    Parameters
    ----------
    mat: numpy.ndarray
        LxL matrix where the mat[i,j]=1 represents paired bases, otherwise 0
    connects: [int], length L
        The i-th base connects to `connects[i-1]`-th base, 1-indexed, 0 for no connection.
    mode: str
        prob, energy, softmax

    Returns
    -------
    scores: [float], length L
        range [0, 1]
    '''
    def get_pred_score(xs, i, mode):
        if i is None:
            if xs.max()==xs.min()==0:
                return 0
            else:
                return 1
        if mode == 'prob':
            prob = xs[i]
            if prob<=0:
                return 0.0
            if prob>1:
                return 1.0
            else:
                return prob
        elif mode == 'energy':
            RT = 2.4788 # kj/mol
            xs = -xs/RT
            x = xs[i]
            sm_exp = (np.exp(xs)).sum()
            return np.exp(x)/sm_exp
        elif mode == 'softmax':
            x = xs[i]
            sm_exp = (np.exp(xs)).sum()
            return np.exp(x)/sm_exp
    scores = []
    for i in range(len(mat)):
        score = get_pred_score(mat[i], None if connects[i]==0 else connects[i]-1, mode)
        scores.append(score)
    return np.array(scores).tolist()


def parse_stockholm(lines):
    '''
    Stockholm format (e.g. Rfam seed file).

    Parameters
    ----------
    lines : [str]
        lines read from file

    Returns
    -------
    ret: dict
        {'headers': {tag: value}, 'seqs': {name: seq}, 'SS': dbn}
    '''
    ret = {'headers': {}, 'seqs': {}, 'SS': ''}
    for line in lines:
        line = line.strip('\r\n')
        parts = [part for part in line.split() if part]
        if line.startswith('#=GF'):
            ret['headers'][parts[1]] = parts[2]
        elif not line.startswith('#'):
            if len(parts) == 2:
                seq_name, seq = parts
                seq_id = '_'.join(seq_name.split('/'))
                if seq_id in ret['seqs']:
                    ret['seqs'][seq_id] += seq
                else:
                    ret['seqs'][seq_id] = seq
        elif line.startswith('#=GC SS_cons'):
            ret['SS'] += parts[2]
    return ret


def read_stockholm(path):
    '''
    Stockholm format (e.g. Rfam seed file).
    Other method to read: scikit-bio
    ``
        from skbio import Protein, TabularMSA
        msa = TabularMSA.read(fp, constructor=Protein)
    ```

    Parameters
    ----------
    path: str
        stockholm file path

    Returns
    -------
    ret: dict
        {'headers': {tag: value}, 'seqs': {name: seq}, 'SS': dbn}
    '''
    with open(path) as fp:
        return parse_stockholm(fp.readlines())


def process_stockholm_SS(seq, dbn):
    '''
        _：表示间隔区域。
        -：表示缺失或未配对的区域。
        :：表示这些位置的碱基保守性较高或成对概率较高。
        ~: 
        The RNA SS letters are taken from WUSS (Washington University Secondary Structure) notation. Matching nested parentheses characters <>, (), [], or {} indicate a basepair. The symbols '.', ',' and ';' indicate unpaired regions. Matched upper and lower case characters from the English alphabet indicate pseudoknot interactions. The 5' nucleotide within the knot should be in uppercase and the 3' nucleotide lowercase.
    '''
    base_set = set('AUGC')
    unpaired_sym = set('.,_-:~')
    seq = seq.upper()
    ref_dbn = ['.' if sym in unpaired_sym else sym for sym in dbn]
    connects = dbn2connects(ref_dbn)
    # firstly, process ref_dbn, unpair unkown base pairs
    for idx, base in enumerate(seq):
        if base not in base_set: # discard
            if connects[idx]!=0: # if have pair, unpair it
                ref_dbn[idx] = ref_dbn[connects[idx]-1] = '.'
    # then, get final seq and dbn
    seq_lst = []
    dbn_lst = []
    for idx, base in enumerate(seq):
        if base in base_set:
            seq_lst.append(base)
            dbn_lst.append(ref_dbn[idx])
    return ''.join(seq_lst), ''.join(dbn_lst)


def compute_confidence(mat1, mat2, scale_k:float=1.5220298, scale_b:float=-0.08571319, ceiling_01:bool=True)->float:
    return cos_row(mat1, mat2, scale_k, scale_b, ceiling_01)


def cos_row(mat1, mat2, scale_k:float=1, scale_b:float=0, ceiling_01:bool=True)->float:
    '''
    Compute confidence index according to contact maps before and after postprocessing.

    Parameters
    ----------
    mat: np.ndarray
        Contact map.

    Returns
    -------
    CI: float
        confidence index
    '''
    CEILING = 0.98
    FLOOR = 0.0
    inner_prod = 0
    norm1 = norm2 = 0
    for arr1, arr2 in zip(mat1, mat2):
        # np.linalg.norm, np.linalg.det
        norm1 += np.dot(arr1, arr1)
        norm2 += np.dot(arr2, arr2)
        inner_prod += np.dot(arr1, arr2)
    if norm1 == 0 and norm2 == 0: # NOTICE! zero vector
        return CEILING # don't use 1
    elif norm1 == 0 or norm2 == 0:
        return FLOOR
    else:
        CI = inner_prod/((norm1*norm2)**0.5)
        # The linear coeffs are for rescale CI to [0, 1], which is the range of F1
        if scale_k and scale_b is not None:
            CI = scale_k*(CI.item()) + scale_b
        if ceiling_01:
            return min(CEILING, max(FLOOR, CI))
        else:
            return CI


def _cal_metric_from_tp(length, pred_p, gt_p, tp, eps=1e-12):
    '''
        pred_p, gt_p, tp can be tensors, can be on GPUs

        accuracy: acc = (TP+TN)/(TP+FP+FN+TN)
        precision: p = TP/(TP+FP)
        recall: r = TP/(TP+FN)
        F1: F1 = 2*p*r / (p+r)
        sensitivity = recal = TPR (true positive rate)
        specificity = TN/(TN+FP)
        YoudenIndex = sen + spe - 1
        false positive rate: FPR = FP/(TN+FP) = 1-spe
        positive predicted value: PPV = precision
        negative predicted value: NPV = TN/(TN+FN)
    '''
    assert 0 <= tp <= pred_p <= length and 0 <= tp <= gt_p <= length, f'length={length}, pred_p={pred_p}, gt_p={gt_p}, tp={tp}'
    fp = pred_p - tp
    fn = gt_p - tp
    tn = length - tp - fp - fn
    metric_dic = {
            'F1': (2*tp + eps)/(2*tp + fp + fn + eps),
            'ACC': (tp + tn + eps)/(tp + fp + fn + tn + eps),
            'MCC': (tp * tn - fp * fn + eps)/(((tp+fp)*(tp+fn)*(tn+fp)*(tn+fn))**0.5+eps),
            'Precision': (tp + eps)/(tp+fp+eps),
            'Recall': (tp + eps)/(tp+fn+eps),
           }
    metric_dic['INF'] = (metric_dic['Precision']*metric_dic['Recall'])**0.5
    # print(length, pred_p, gt_p, tp, fp, tn, fn)
    # print(metric_dic)
    for k, v in metric_dic.items():
        if k in ['F1', 'ACC', 'Precision', 'Recall', 'INF']:
            assert 0 <= v <= 1, f'{k}={v}'
        elif k == 'MCC':
            assert -1 <= v <= 1, f'{k}={v}'
        else:
            raise Exception(f'Unkonwn metric name: {k}')
    return metric_dic


def cal_metric(pred_mat, gt_mat):
    '''
        pred_mat, gt_mat: LxL, 0, 1 valued
    '''
    L = len(pred_mat)
    if L!=len(gt_mat):
        raise Exception(f'[Error]: lengthes dismatch: pred {L}!= gt {len(gt_mat)}')
    if L==0:
        return 1
    assert len(pred_mat[0])==len(gt_mat[0])==L, f'[Error] Contact map shape inconsistent: {L}!={pred_mat[0]}, {gt_mat[0]}'
    pred_p = gt_p = tp = 0 # predpair, gtpair, paired
    for i in range(L):
        for j in range(L):
            pred = pred_mat[i, j]
            gt = gt_mat[i, j]
            if gt!=0:
                gt_p +=1
            if pred!=0:
                pred_p +=1
                if pred==gt:
                    tp +=1
    return _cal_metric_from_tp(L*L, pred_p, gt_p, tp)


def cal_metric_pairwise(pred_pairs:[int], gt_pairs:[int]):
    '''
        # deprecated
        pred_pairs, gt_pairs: connections, 1-indexed
        return: MCC, INF, F1, precision, recall
    '''
    length = len(pred_pairs)
    if length!=len(gt_pairs):
        raise Exception(f'[Error]: lengthes dismatch: pred {length}!= gt {len(gt_pairs)}')
    pred_p = gt_p = tp = 0 # predpair, gtpair, paired
    for pred, gt in zip(pred_pairs, gt_pairs):
        if gt!=0:
            gt_p +=1
        if pred!=0:
            pred_p +=1
            if pred==gt:
                tp +=1
    return _cal_metric_from_tp(length, pred_p, gt_p, tp)
